/*
*
* Author:              Vadim V. Sinitsyn
* E-mail:              sinvv@asplinux.ru
*
* $Header: /home/cvs/ASPcomplete/linux/include/linux/dq_asp_def.h,v 1.3 2000/10/03 11:49:36 sinvv Exp $
*
* Remarks:
*
* Last correct: 18/Oct/00
*
*/
#ifndef	_DQ_ASP_DEF_H_
#define	_DQ_ASP_DEF_H_

#include <asm/errno.h>
#include <asm/types.h>

#undef in
#undef out
#undef inout

#define in
#define out
#define inout

/*
* Current ASP Disk Quota data base layout
*	--------------------------
*	| MAGIC NUMBER   8 bytes |
*	--------------------------
*	| VERSION NUMBER 4 bytes |
*	--------------------------
*	| COUNTER        4 bytes |
*	--------------------------
*	|  Record1               |
*	~~~~~~~~~~~~~~~~~~~~~~~~~~
*	|  .....                 |
*	~~~~~~~~~~~~~~~~~~~~~~~~~~
*	|  RecordN               |
*	--------------------------
*
* Counter = N
* Record = 13x4 unsigned int=52 bytes
*
* Record: 
* veid,luid,uid,gid,flags,
* unit_current,unit_soft,unit_hard,unit_expiration,
* inode_current,inode_soft,inode_hard,inode_expiration
*
* Empty base = 8+4+4=16 bytes
*
*/

#define DQ_ASP_VERSION      2

//#define DQ_ASP_DEBUG
//#define DQ_ASP_DEBUG1

#ifdef DQ_ASP_DEBUG
#define DQ_DEBUG(format, args...)    printk(__FUNCTION__":" format, ## args);
#else
#define DQ_DEBUG(format, args...)
#endif

#ifdef DQ_ASP_DEBUG1
#define DQ_DEBUG1(format, args...)    printk(__FUNCTION__":" format, ## args);
#else
#define DQ_DEBUG1(format, args...)
#endif

#define DBG

#ifdef DBG
#define dprint	printf
#else
#define dprint
#endif

typedef struct dq_blk {
   	__u32	veid;            // Virtual Environment ID
	uid_t	luid;            // Login User ID
	uid_t	uid;             // User ID
	gid_t	gid;             // Group ID
	__u32	flags;           // Flags for quota description
	__u32	unit_current;    // Current unit value
	__u32	unit_soft;       // Soft unit limit
	__u32	unit_hard;       // Hard unit limit
	time_t	cur_unit_time;   // Current unit time (countdown counter for soft unit limit) 
	time_t	unit_expiration; // Expiration time (for soft unit time)
	__u32	inode_current;   // Current inode value
	__u32	inode_soft;      // Soft inode limit
	__u32	inode_hard;      // Hard inode limit
	time_t  cur_inode_time;  // Current inode time (countdown counter for soft inode limit) 
	time_t	inode_expiration;// Expiration time (for soft inode time)
} dq_blk;

typedef struct ldq_blk {	// list dq_blk
	struct ldq_blk	*next;	
	__u32			offset;	// Offset in quota data base
	struct file		*fp;	// Ptr to open quota file
	struct dq_blk	dq;
} ldq_blk;

#define DQ_ASP_SIZE			sizeof(struct dq_blk)

#define MAGIC			"xunilpsa"						// Magic file prefix
#define MAGICLEN		8								// Magic number len
#define DQB_HDRSIZE		(2*sizeof(__u32)+MAGICLEN)		// Header size for qdb
#define DQB_VERSION		2								// Quota Data Base ver
#define ROOT_UID		0								// Root uid
#define VEID_0			0								// Main veid
#define NOTRUN_VE		-1								// VE not rus now
#define DQB_RIGHT		(S_IFREG | S_IRUSR | S_IWUSR)	// 0100600
#define DEFAULT_DQB		"/var/quota/dqb"                // Default value to data base ASP Disk Quota
#define PATH_DQB_ENV    "DQB"                           // Environment variable for path to data base ASP Disk Quota
#define DEFAULT_EXPIRATION	(7*24*60*60)				// 1 week
#define PATH_SIZE		4096							// Size for name - 1 page

// DQ flags
#define DQ_ASP_VEID			0		// VEID quota (by veid) groot
#define DQ_ASP_UID			1		// UID quota (by users)	groot,veroot
#define DQ_ASP_GID			2		// GID quota (by groups) groot,veroot
#define DQ_ASP_MAXTYPE		3		// Number of quota types
#define DQ_ASP_VEGUID		7
#define DQ_ASP_VEROOT		0x08	// Was set veroot (private quotas)
#define DQ_ASP_DISABLE		0x10	// Disable some quota
#define DQ_ASP_REMOVE		0x20	// Remove quota from DQB
#define DQ_ASP_DIRTY		0x40	// Dirty flag (quota was changed)

#define DQ_ASP_FORMAT_LINE		1
#define DQ_ASP_FORMAT_COLUMN	2
#define DQ_ASP_FORMAT_UNIT_KB	4
#define DQ_ASP_FORMAT_UNIT_MB	8

#define DQ_ASP_OFF				0
#define DQ_ASP_ON				1

#define DQ_ASP_BLOCK			512
#define KB					1024
#define MB					(1024*1024)
#define MAX_ALLOC_SIZE		128*KB

typedef char				path[PATH_SIZE];

#define DQ_ASP_CMD_GET_VERSION		0
#define DQ_ASP_CMD_ENABLE		    1
#define DQ_ASP_CMD_DISABLE		    2
#define DQ_ASP_CMD_SYNC			    3
#define DQ_ASP_CMD_GET			    4
#define DQ_ASP_CMD_PUT			    5
#define DQ_ASP_CMD_REMOVE		    6
#define DQ_ASP_CMD_OFF			    7
#define DQ_ASP_CMD_GETSTAT		    8
#define DQ_ASP_CMD_RESETSTAT	    9
#define DQ_ASP_CMD_GETSIZE	        10
#define DQ_ASP_CMD_READFILE		    11
#define DQ_ASP_CMD_WRITEFILE	    12
#define DQ_ASP_CMD_TEST			    13
#define DQ_ASP_CMD_GET_SD_SIZE      14
#define DQ_ASP_CMD_GET_INODE        15
#define DQ_ASP_CMD_GET_BLOCK        16
#define DQ_ASP_CMD_GET_BLOCKSIZE    17
#define DQ_ASP_CMD_GET_KDEV         18


typedef struct dq_stat {
	__u32		all;			// Whole quantity
	__u32		disable;
	__u32		veroot;			// How much restrictions set veroots
	__u32		type[DQ_ASP_MAXTYPE];
} dq_stat;

// flag primitives

#define DQ_ASP_CMP_TYPE(p,q)		(((p)->flags & DQ_ASP_VEGUID)==((q)->flags & DQ_ASP_VEGUID))

#define DQ_ASP_IS_REMOVE(p)			((p) & DQ_ASP_REMOVE)
#define DQ_ASP_IS_DIRTY(p)			((p) & DQ_ASP_DIRTY)
#define DQ_ASP_IS_DISABLE(p)		((p) & DQ_ASP_DISABLE)
#define DQ_ASP_IS_VEROOT(p)			((p) & DQ_ASP_VEROOT)
#define DQ_ASP_IS_VEID(p)			(DQ_ASP_VEID == ((p) & DQ_ASP_VEGUID))
#define DQ_ASP_IS_UID(p)			(DQ_ASP_UID == ((p) & DQ_ASP_VEGUID))
#define DQ_ASP_IS_GID(p)			(DQ_ASP_GID == ((p) & DQ_ASP_VEGUID))

#define DQ_ASP_EQUAL(p,q)	\
			(((p)->veid == (q)->veid) &&					\
			  ((p)->uid == (q)->uid) &&						\
			  ((p)->gid == (q)->gid))

#define DQ_ASP_COPY(p,q)	\
		(p)->veid			= (q)->veid;				\
		(p)->luid			= (q)->luid;				\
		(p)->uid			= (q)->uid;					\
		(p)->gid			= (q)->gid;					\
		(p)->flags			= (q)->flags;				\
		(p)->unit_current	= (q)->unit_current;		\
		(p)->unit_soft		= (q)->unit_soft;			\
		(p)->unit_hard		= (q)->unit_hard;			\
		(p)->cur_unit_time	= (q)->cur_unit_time;		\
		(p)->unit_expiration= (q)->unit_expiration;		\
		(p)->inode_soft		= (q)->inode_soft;			\
		(p)->inode_hard		= (q)->inode_hard;			\
		(p)->inode_current	= (q)->inode_current;		\
		(p)->inode_expiration = (q)->inode_expiration;	\
		(p)->cur_inode_time	= (q)->cur_inode_time;

#define DQ_ASP_GET_USER_COPY(p,q)	\
		get_user((p)->veid,&((q)->veid));  				\
		get_user((p)->luid,&((q)->luid));				   \
		get_user((p)->uid,&((q)->uid));					\
		get_user((p)->gid,&((q)->gid));					\
		get_user((p)->flags,&((q)->flags));				\
		get_user((p)->unit_current,&((q)->unit_current));	\
		get_user((p)->unit_soft,&((q)->unit_soft));		\
		get_user((p)->unit_hard,&((q)->unit_hard));		\
		get_user((p)->cur_unit_time,&((q)->cur_unit_time));	\
		get_user((p)->unit_expiration,&((q)->unit_expiration));\
		get_user((p)->inode_soft,&((q)->inode_soft));		\
		get_user((p)->inode_hard,&((q)->inode_hard));		\
		get_user((p)->inode_current,&((q)->inode_current));	\
		get_user((p)->inode_expiration,&((q)->inode_expiration));	\
		get_user((p)->cur_inode_time,&((q)->cur_inode_time));

#define DQ_ASP_PUT_USER_COPY(q,p)	\
		put_user((p)->veid,&((q)->veid));  				\
		put_user((p)->luid,&((q)->luid));				\
		put_user((p)->uid,&((q)->uid));					\
		put_user((p)->gid,&((q)->gid));					\
		put_user((p)->flags,&((q)->flags));				\
		put_user((p)->unit_current,&((q)->unit_current));	\
		put_user((p)->unit_soft,&((q)->unit_soft));		\
		put_user((p)->unit_hard,&((q)->unit_hard));		\
		put_user((p)->cur_unit_time,&((q)->cur_unit_time));	\
		put_user((p)->unit_expiration,&((q)->unit_expiration));\
		put_user((p)->inode_soft,&((q)->inode_soft));		\
		put_user((p)->inode_hard,&((q)->inode_hard));		\
		put_user((p)->inode_current,&((q)->inode_current));	\
		put_user((p)->inode_expiration,&((q)->inode_expiration));	\
		put_user((p)->cur_inode_time,&((q)->cur_inode_time));


#define DQ_ASP_LCOPY(p,q)	\
		(p)->dq.veid			= (q)->dq.veid;				\
		(p)->dq.luid			= (q)->dq.luid;				\
		(p)->dq.uid			    = (q)->dq.uid;				\
		(p)->dq.gid			    = (q)->dq.gid;				\
		(p)->dq.flags			= (q)->dq.flags;			\
		(p)->dq.unit_current	= (q)->dq.unit_current;		\
		(p)->dq.unit_soft		= (q)->dq.unit_soft;		\
		(p)->dq.unit_hard		= (q)->dq.unit_hard;		\
		(p)->dq.cur_unit_time	= (q)->dq.cur_unit_time;	\
		(p)->dq.unit_expiration= (q)->dq.unit_expiration;	\
		(p)->dq.inode_soft		= (q)->dq.inode_soft;		\
		(p)->dq.inode_hard		= (q)->dq.inode_hard;		\
		(p)->dq.inode_current	= (q)->dq.inode_current;	\
		(p)->dq.inode_expiration = (q)->dq.inode_expiration;\
		(p)->dq.cur_inode_time	= (q)->dq.cur_inode_time;   \
        (p)->fp                 = (q)->fp;                  \
        (p)->offset             = (q)->offset;


// Set soft limit

#define DQ_ASP_SOFT_LIMIT_INODE(p)	\
	if (0 != (p)->inode_soft) {							\
		if ((p)->inode_current > (p)->inode_soft) {		\
			if (0 == (p)->cur_inode_time) {				\
				(p)->cur_inode_time = CURRENT_TIME + (p)->inode_expiration; \
				printk("start inode soft limit\n");		\
			}											\
			else if (CURRENT_TIME > (p)->cur_inode_time) { \
				printk("inode soft limit: time expired\n"); \
			}											\
		}												\
		else (p)->cur_inode_time = (time_t)(0);			\
	}													\
	else (p)->cur_inode_time = (time_t)(0);

#define DQ_ASP_SOFT_LIMIT_UNIT(p)		\
	if (0 != (p)->unit_soft) {							\
		if ((p)->unit_current > (p)->unit_soft) {		\
			if (0 == (p)->cur_unit_time) {				\
				(p)->cur_unit_time = CURRENT_TIME + (p)->unit_expiration; \
				printk("start unit soft limit\n");		\
			}											\
			else if (CURRENT_TIME > (p)->cur_unit_time) { \
				printk("unit soft limit: time expired\n"); \
			}											\
		}												\
		else (p)->cur_unit_time = (time_t)(0);			\
	}													\
	else (p)->cur_unit_time = (time_t)(0);



// dq_stat primitives

#define DQ_ASP_INC_STAT(stat,ldq)		\
			(stat)->all++;													\
			if (DQ_ASP_MAXTYPE > ((ldq)->dq.flags & DQ_ASP_VEGUID)) {				\
				(stat)->type[(ldq)->dq.flags & DQ_ASP_VEGUID]++;				\
			}																\
			if (DQ_ASP_IS_DISABLE((ldq)->dq.flags)) (stat)->disable++;			\
			if (DQ_ASP_IS_VEROOT((ldq)->dq.flags)) (stat)->veroot++;

#define DQ_ASP_DEC_STAT(stat,ldq)		\
			if (0 < (stat)->all) (stat)->all--;								\
			if (DQ_ASP_MAXTYPE > ((ldq)->dq.flags & DQ_ASP_VEGUID)) {				\
				if (0 < (stat)->type[(ldq)->dq.flags & DQ_ASP_VEGUID])			\
					(stat)->type[(ldq)->dq.flags & DQ_ASP_VEGUID]--;			\
			};																\
			if (DQ_ASP_IS_DISABLE((ldq)->dq.flags)) {							\
				if (0 < (stat)->disable) (stat)->disable--;					\
			};																\
			if (DQ_ASP_IS_VEROOT((ldq)->dq.flags))								\
				if (0 < (stat)->veroot) (stat)->veroot--;

#define DQ_ASP_SET_VEGUID(dq)		\
			(dq)->veid	= current->envid->veid;	\
			(dq)->luid	= current->luid;		\
			(dq)->uid	= current->uid;			\
			(dq)->gid	= current->gid;
		

#endif  // _DQ_ASP_DEF_H_



