#ifndef _LINUX_VE_H
#define _LINUX_VE_H

#include <linux/config.h>

#ifdef CONFIG_VE

#include <linux/types.h>
#include <linux/capability.h>
#include <linux/utsname.h>

#define VE_CREATE 	1	/* Create VE, VE_ENTER added automatically */
#define VE_EXCLUSIVE	2	/* Fail if exists */
#define VE_ENTER	4	/* Enter existing VE */
#define VE_TEST		8	/* Test if VE exists */

typedef __u32 envid_t;

#ifdef __KERNEL__
struct tty_driver;
struct task_struct;
struct new_utsname;
struct devpts_sb_info;

struct pty_struct {
	int	magic;
	wait_queue_head_t open_wait;
};

struct ve_struct 
{
	envid_t			veid;
	struct task_struct	*init_entry;
	kernel_cap_t		cap_default;
	atomic_t		pcounter;
	struct semaphore	init_exit_guard;
	struct new_utsname	_system_utsname;

/* VE's root */
	struct vfsmount 	*fs_rootmnt;
	struct dentry 		*fs_root;
	
/* BSD pty's */
	int 			pty_refcount;
	struct tty_driver       *pty_driver;
	struct tty_driver       *pty_slave_driver;

#ifdef CONFIG_UNIX98_PTYS
	struct tty_driver	*ptm_driver[UNIX98_NR_MAJORS];
	struct tty_driver	*pts_driver[UNIX98_NR_MAJORS];
#endif
#ifdef CONFIG_DEVPTS_FS
	struct devpts_sb_info	*pts_sb_info;
#endif

/* Bound virtual network interface */
	u32			ip;
	
	struct ve_struct	*prev;
	struct ve_struct	*next;

/* saved idle / cpu stats */
	unsigned long 		start_jiffies;
	unsigned long		idle;
	unsigned int		per_cpu_user[NR_CPUS];
	unsigned int		per_cpu_nice[NR_CPUS];
	unsigned int		per_cpu_system[NR_CPUS];
};

#define ve_offsetof(arg)       	((void*)offsetof( struct ve_struct,_##arg ))

unsigned char get_device_perms_ve( envid_t envid, int type, kdev_t dev );
void do_env_cleanup( struct ve_struct *envid );
void do_mark_env_to_down( struct ve_struct *ptr );

#ifdef CONFIG_VE_NET
extern struct ve_struct *get_ve_by_ip(u32 addr);
#endif
#ifdef CONFIG_VE
#define system_utsname (is_ve_initialized() ? (current->envid->_system_utsname) : _system_utsname)
#endif
#endif /* __KERNEL__ */

#define VELINK 12

#define VE_CONF_MSG	0x11
#define VE_REGISTER	0x12

struct ve_conf_request
{
	envid_t veid;
	ulong   id;
	int	answer;
	char	data[256];
};

#else  /* CONFIG_VE */

#define ve_offsetof(arg)        (&arg)

#endif /* CONFIG_VE */
#endif /* _LINUX_VE_H */
