/*
*
* Author:              Vadim V. Sinitsyn
* E-mail:              sinvv@asplinux.ru
*
* $Header: /home/cvs/ASPcomplete/linux/kernel/dq_asp.c,v 1.39 2000/10/10 07:54:57 sinvv Exp $
*
* Remarks:
*
* Last correct: 16/Oct/00
*
*/
#include <linux/sched.h>
#include <linux/delay.h>
#include <linux/capability.h>
#include <linux/ve.h>
#include <linux/smp_lock.h>
#include <linux/init.h>
#include <linux/list.h>

#include <linux/errno.h>
#include <linux/unistd.h>
#include <linux/malloc.h>
#include <linux/sys.h>
#include <linux/kdev_t.h>
#include <linux/tty_driver.h>
#include <linux/netdevice.h>
#include <linux/utsname.h>
#include <linux/proc_fs.h>
#include <asm/uaccess.h>

#include <linux/reiserfs_fs.h>

#include <linux/dq_asp_def.h>
#include <linux/dq_asp.h>

//#define DQ_TEST

static struct ldq_blk	*dq_hash[DQ_ASP_HASH_SIZE];
static spinlock_t		dq_spinlock[DQ_ASP_HASH_SIZE];
//static struct semaphore dq_sem[DQ_ASP_HASH_SIZE];
static kmem_cache_t		*dq_cachep;
static struct dq_stat	dqstat;
static int	dq_asp_get_info(char *buffer, off_t offset, int length);

// Extern functions
extern int dq_lock_file(struct file *fp, __u32 type, __u32 start, __u32 len);

// Static functions
static inline u32 dq_hash_func(in struct dq_blk *dq);
static int		dq_insert_hash(in struct ldq_blk *ldq);
static int      dq_flags_hash(in struct dq_blk *dq, in u32 flags, in u32 action);
static int      dq_off_hash(in struct dq_blk *dq);
static int		dq_flags_type(in u32 type, in u32 flags, in u32 action);
static int		dq_get_hash(inout struct dq_blk *dqb);
static int		dq_flush_hash(in struct dq_blk *dqb);
static int		dq_mark_remove_hash(in struct dq_blk *dqb);
static int		dq_off(in u32 veid, in int type);
static u32		calc_flags(in u32 flag);
static struct file *dq_open_file(in const char *fname, in u32 mode, in u32 access);
static int		dq_close_file(in struct file *fp);
static int		dq_write_file(in struct file *fp, in char *buf, in u32 len);
static int		dq_flush_file(in struct ldq_blk *ldq);
static int		dq_check_inode_limit(in struct dq_blk *dqb, in u32 inodes);
static int		dq_check_unit_limit(in struct dq_blk *dqb, in u32 units, in u32 prealloc);

int				dq_alloc_unit(in u32 units, in u32 prealloc);
int				dq_check_unit(in u32 units);
int				dq_alloc_inode(in u32 inodes);
void			dq_free_unit(in u32 units);
void			dq_free_inode(in u32 inodes);
void			dq_sync(in u32 force);

// Misc print functions
#ifdef DQ_TEST
static void pr_dentry(struct dentry *de);
static void pr_stat_data(struct stat_data *st);
static void pr_de_head(struct reiserfs_de_head *deh);
static void pr_dir_entry(struct reiserfs_dir_entry *de);
//static void pr_essential_constants(void);
static void pr_inode(struct inode *inode);
static void pr_cpu_key(char *s, struct cpu_key  *key);
static void pr_ih(char *s, struct item_head *ih);
static void pr_bufh(char *s, struct buffer_head *bufh);
static void pr_blockh(char *s, struct block_head *blockh);
static int look_for(struct inode *inode, int offset, int type);
static void look_for_ok(struct inode *inode, int offset, int type, int cnt);
static void get_type(char *s, int type);
#endif

static int dq_test_get_sd_size(in char *fpath, out int *sd_size);
static int dq_test_get_inode(in char *fpath, out int *inode);
static int dq_test_get_block(in u32 dev, in int block_number, out void *buf, in int buflen);
static int dq_test_get_blocksize(in char *fpath, out int *blocksize);
static int dq_test_get_kdev(in char *fpath, out int *kdev, out char *mnt, out char *dev_name);
static int fpath2inode(in char *fpath, out struct inode *inode);
static void dq_free(struct ldq_blk *p);

#ifdef DQ_TEST
static int try_key(struct inode *inode, 
                   struct path  *path,
                    int offset, 
                    int type, 
                    struct item_head **ih, 
                    struct buffer_head **bh);
#endif
// Syscalls
asmlinkage int	sys_dq_asp(u32 cmd, u32 n, u32 addr1, u32 addr2);

static int dq_getsize(in char *fpath, out int *n);

#define SET_SPINLOCK(x)							\
	spin_lock(&dq_spinlock[x]);					\

#define RELEASE_SPINLOCK(x)						\
	if (-1 != x) spin_unlock(&dq_spinlock[x]);	\

#define SET_SEMAPHORE(x)						\
	down_interruptible(&dq_sem[x]);				
	
#define RELEASE_SEMAPHORE(x)					\
    if (-1 != x) up(&dq_sem[x]);                

//#define SET_PROTECT(x)              SET_SEMAPHORE(x)
//#define RELEASE_PROTECT(x)          RELEASE_SEMAPHORE(x)
#define SET_PROTECT(x)				SET_SPINLOCK(x)
#define RELEASE_PROTECT(x)			RELEASE_SPINLOCK(x)

int pflag1=0;
int fp_cnt=0;

/*
* Run in main 
*/
void __init dq_asp_init_hash(void) {

	int	i;

	printk(KERN_NOTICE "ASP_DQ: ASP Disk Quotas version %d initialized\n", DQ_ASP_VERSION);

	dq_cachep = kmem_cache_create("dq_asp", sizeof(struct ldq_blk),
					 0, SLAB_HWCACHE_ALIGN, NULL, NULL);

	if (0 == dq_cachep)
		panic("Cannot create dquot SLAB cache\n");

	memset((void *)dq_hash, 0, DQ_ASP_HASH_SIZE * sizeof(struct ldq_blk *));
	memset((void *)&dqstat, 0, sizeof(dq_stat));

	for (i = 0; i < DQ_ASP_HASH_SIZE; i++) {
		spin_lock_init(&dq_spinlock[i]);
        //sema_init(&dq_sem[i], 1);
	}
}
/* 
 * For PROC filesystem output
 */
int dq_asp_read_proc(char *page, char **start, off_t off, int count,
			int *eof, void *data)
{
	int len;

	len = dq_asp_get_info(page, off, count);
	*start = page;
	if (len>count) len = count;
	if (len<0) len = 0;
	*eof = len < count;
	page[len]=0;
	return len;
}
/*
 * Get information for PROC filesystem output
 */
static int dq_asp_get_info(char *page, off_t offset, int length)
{
	char 			*buffer;
	int				len=0;
	struct dq_blk	dq, dqv = { 0 }, dqu = { 0 }, dqg = { 0 };
	
	DQ_ASP_SET_VEGUID(&dq);
	*page = 0;
	// veid type
	dqv.veid = dq.veid; dqv.uid = dqv.gid = 0; dqv.flags = DQ_ASP_VEID;
	if (0 != dqv.veid) dq_get_hash(&dqv);
	// uid type
	dqu.veid = dq.veid; dqu.uid = dq.uid; dqu.gid = 0; dqu.flags = DQ_ASP_UID;
	if (0 != dqu.uid) dq_get_hash(&dqu);
	// gid type
	dqg.veid = dq.veid; dqg.gid = dq.gid; dqg.uid = 0; dqg.flags = DQ_ASP_GID;
	if (0 != dqg.gid) dq_get_hash(&dqg);
	
	buffer = kmalloc(4096, GFP_ATOMIC);
	if (0 == buffer) {
		printk("dq_asp_get_info: can't allocate memory\n");
		return 0;
	}
	len = sprintf(buffer, "%d %d %d\n\
%d %d %d %d %d %d %d %d %d %d\n\
%d %d %d %d %d %d %d %d %d %d\n\
%d %d %d %d %d %d %d %d %d %d\n", 
		dq.veid, dq.uid, dq.gid,
		// veid type
		dqv.unit_current, dqv.unit_soft, dqv.unit_hard,
		(int)dqv.cur_unit_time, (int)dqv.unit_expiration,
		dqv.inode_current, dqv.inode_soft, dqv.inode_hard,
		(int)dqv.cur_inode_time, (int)dqv.inode_expiration,
		// uid type
		dqu.unit_current, dqu.unit_soft, dqu.unit_hard,
		(int)dqu.cur_unit_time, (int)dqu.unit_expiration,
		dqu.inode_current, dqu.inode_soft, dqu.inode_hard,
		(int)dqu.cur_inode_time, (int)dqu.inode_expiration,
		// gid type
		dqg.unit_current, dqg.unit_soft, dqg.unit_hard,
		(int)dqg.cur_unit_time, (int)dqg.unit_expiration,
		dqg.inode_current, dqg.inode_soft, dqg.inode_hard,
		(int)dqg.cur_inode_time, (int)dqg.inode_expiration);

	if (offset >= len) {
		kfree(buffer);
		return 0;
	}
	
	if (len > (offset+length)) len = length;
	else len -= offset;
	strncat(page, &buffer[offset], len);
	kfree(buffer);
	return len;
}

/*
 * Hash function for fast lookup for list
 */
static u32 hash_data[16] = {
	0x67452301, 0xefcdab89, 0x98badcfe, 0x10325476,
	0xc3d2e1f0, 0x5a827999, 0x6ed9eba1, 0x8f1bbcdc,
	0x29173786, 0x24650664, 0x37172713, 0x210c6994,
	0x9348605a, 0x38d01683, 0x363f4896, 0x3894e248
};

static inline u32 dq_hash_func(in struct dq_blk *dq) {

	return (dq->veid^hash_data[(dq->uid>>3)&0x0f]^hash_data[(dq->gid<<1)&0x0f]) % DQ_ASP_HASH_SIZE;
}
/*
 * Insert quota record to hash table
 * Parameters:
 *    in struct ldq_blk *ldq	- structure for insert
 * Return:
 *    0		- Ok
 */
static int dq_insert_hash(in struct ldq_blk *ldq) {
	
	u32				i;
	struct ldq_blk	*q, *p;

DQ_DEBUG("start\n");
	// Calc hash value
	i = dq_hash_func(&ldq->dq);
	// Set protect
    SET_PROTECT(i);
	// Try find at the same record
	q = dq_hash[i];
	while ( q ) {
		if (DQ_ASP_CMP_TYPE(&q->dq,&ldq->dq)) {
			if (DQ_ASP_EQUAL(&q->dq,&ldq->dq)) {
                // Close more opened file
                if (0 != q->fp) dq_close_file(q->fp);
				DQ_ASP_LCOPY(q,ldq);
				
				// Check soft limits
				DQ_ASP_SOFT_LIMIT_INODE(&q->dq);
				DQ_ASP_SOFT_LIMIT_UNIT(&q->dq);
	            // Release protect
                RELEASE_PROTECT(i);
				return 0;
			}
		}
		q = q->next;
	}

    p = kmem_cache_alloc(dq_cachep, SLAB_KERNEL);
    if (NULL == p) return ENOMEM;
	DQ_ASP_LCOPY(p,ldq);
	
	q = dq_hash[i];
	// If q == 0 (empty entry) - no problems
	dq_hash[i] = p;
	p->next = q;
	// Check soft limits
	DQ_ASP_SOFT_LIMIT_INODE(&p->dq);
	DQ_ASP_SOFT_LIMIT_UNIT(&p->dq);
	// Release protect
    RELEASE_PROTECT(i);
	// Change statistics
	DQ_ASP_INC_STAT(&dqstat, p);
	return 0;
}
/*
 * Set/reset quota record flag to hash table
 * Parameters:
 *    in struct dq_blk *dq	- structure for change flags
 *    in u32 flags			- flags for change
 *    in u32 action			- DQ_ASP_ON/DQ_ASP_OFF
 * Return:
 *    0			- Successful set/reset
 *    -1		- Can't  find to hash table
 *    EINVAL	- Bad input params
 */
static int dq_flags_hash(in struct dq_blk *dq, in u32 flags, in u32 action) {
	
	u32				i;
	struct ldq_blk	*cur, *first, *last;

	// Check input params
	if (NULL == dq) return EINVAL;
	// Check flags
	if (1 != calc_flags(flags)) return EINVAL;
	// We can't reset REMOVE flag
	if ((DQ_ASP_IS_REMOVE(flags)) && (DQ_ASP_OFF == action)) return EINVAL;
	// Calc hash value
	i = dq_hash_func(dq);
	// Set protect
    SET_PROTECT(i);

	first = last = cur = dq_hash[i];

	while ( cur ) {
		if (DQ_ASP_CMP_TYPE(&(cur->dq), dq)) {
			if (DQ_ASP_EQUAL(&(cur->dq), dq)) {
				// Set
				if (DQ_ASP_ON == action) {
					// Transition: 0->1
					if (0 == (cur->dq.flags & flags)) {
						// If previous state OFF
						cur->dq.flags |= flags | DQ_ASP_DIRTY;
						// Change statistics	
						if (DQ_ASP_IS_DISABLE(flags)) dqstat.disable++;
						else if (DQ_ASP_IS_VEROOT(flags)) dqstat.veroot++;
						else if (DQ_ASP_IS_REMOVE(flags)) {
                            // Remove quota from list
				            if (first == cur) {	// Head of list
					            dq_hash[i] = cur->next;
				            }
				            else last->next = cur->next;
                            // Release data base file
				            dq_close_file(cur->fp);
				            // Dec statistics
				            DQ_ASP_DEC_STAT(&dqstat, cur);
				            // Release memory
				            kmem_cache_free(dq_cachep, (void *)cur);
                        }
					}
				}
				else { // Reset
					// Transition: 1->0
					if (cur->dq.flags & flags) {
						cur->dq.flags &= ~flags;
						cur->dq.flags |= DQ_ASP_DIRTY;
						// Change statistics
						if (DQ_ASP_IS_DISABLE(flags)) {
							if (0 < dqstat.disable) dqstat.disable--;
						}
						else if (DQ_ASP_IS_VEROOT(flags)) {
							if (0 < dqstat.veroot) dqstat.veroot--;
						}
					}
				}
				// Release protect
                RELEASE_PROTECT(i);
				return 0;
			}
		}
        last = cur;
		cur = cur->next;
	}
	// Release protect
    RELEASE_PROTECT(i);
	return -1;
}
/*
 * Quota record OFF to hash table
 * Parameters:
 *    in struct dq_blk *dq	- structure for OFF
 * Return:
 *    0			- Successful set/reset
 *    -1		- Can't  find to hash table
 *    EINVAL	- Bad input params
 */
static int dq_off_hash(in struct dq_blk *dq) {
	
	u32				i;
	struct ldq_blk	*cur, *first, *last;

	// Check input params
	if (NULL == dq) return EINVAL;
	// Calc hash value
	i = dq_hash_func(dq);
	// Set protect
    SET_PROTECT(i);

	first = last = cur = dq_hash[i];

	while ( cur ) {
		if (DQ_ASP_CMP_TYPE(&(cur->dq), dq)) {
			if (DQ_ASP_EQUAL(&(cur->dq), dq)) {
				// Remove quota from list
				if (first == cur) {	// Head of list
					dq_hash[i] = cur->next;
				}
				else last->next = cur->next;
				// Release protect
                RELEASE_PROTECT(i);
		        // Flush to disk
		        dq_flush_file(cur);
                // Release data base file
				dq_close_file(cur->fp);
				// Dec statistics
				DQ_ASP_DEC_STAT(&dqstat, cur);
				// Release memory
				kmem_cache_free(dq_cachep, (void *)cur);
				return 0;
			}
        }
        last = cur;
		cur = cur->next;
	}
	// Release protect
    RELEASE_PROTECT(i);
	return -1;
}
/*
 * Set/reset quota record flag to hash table for appropriate type
 * (VEID, UID, GID)
 * Parameters:
 *    in u32 type			- Type quota for change
 *    in u32 flags          - Flags for change
 *    in u32 action			- DQ_ASP_ON/DQ_ASP_OFF
 * Return:
 *    0			- Successful set/reset
 *    -1		- Can't  from hash table
 *    EINVAL	- Bad input params
 */
static int dq_flags_type(in u32 type, in u32 flags, in u32 action) {
	
	u32				i;
	struct ldq_blk	*cur;

	// Check type
	if (DQ_ASP_MAXTYPE <= type) return EINVAL;
	// Check input params
	if (1 != calc_flags(flags)) return EINVAL;
	if (flags <= DQ_ASP_VEGUID) return EINVAL;
	// We can't reset REMOVE flag 
	if ((DQ_ASP_IS_REMOVE(flags)) && (DQ_ASP_OFF == action)) return EINVAL;

	for (i = 0; i < DQ_ASP_HASH_SIZE; i++) {
		// Set protect
        SET_PROTECT(i);

		cur = dq_hash[i];

		while ( cur ) {
			if ((cur->dq.flags & DQ_ASP_VEGUID) == type)	{
				if (DQ_ASP_ON == action) { // set
					// Transition: 0->1
					if (0 == (cur->dq.flags & flags)) {
						cur->dq.flags |= flags | DQ_ASP_DIRTY;
						if (DQ_ASP_IS_DISABLE(flags)) dqstat.disable++;
						else if (DQ_ASP_IS_VEROOT(flags)) dqstat.veroot++;
						else if (DQ_ASP_IS_REMOVE(flags)) {
							dq_close_file(cur->fp);
							DQ_ASP_DEC_STAT(&dqstat, cur);
						}
					}
				}
				else { // reset
					// Transition: 1->0
					if (cur->dq.flags & flags) {
						cur->dq.flags &= ~flags;
						cur->dq.flags |= DQ_ASP_DIRTY;
						// Change statistics
						if (DQ_ASP_IS_DISABLE(flags)) {
							if (0 < dqstat.disable) dqstat.disable--;
						}
						else if (DQ_ASP_IS_VEROOT(flags)) {
							if (0 < dqstat.veroot) dqstat.veroot--;
						}
					}
				}
				
			}
			cur = cur->next;
		}
		// Release protect
        RELEASE_PROTECT(i);
	}
	return 0;
}
/*
 * Get quota record from hash table
 * Parameters:
 *    inout struct dq_blk *dqb	- structure for find (set veid/luid/uid/gid/flags)
 *                                we retrieve other fields
 * Return:
 *    0			- Ok
 *	  -1		- Not found	  
 *    EINVAL	- Bad input params
 */
static int dq_get_hash(inout struct dq_blk *dqb) {
	
	u32				i;
	struct ldq_blk	*cur;

DQ_DEBUG("start\n");
	/* Check input params */
	if (0 == dqb) return EINVAL;
	/* Calc hash value  */
	i = dq_hash_func(dqb);
	/* Set protect */
    SET_PROTECT(i);
	/* Set in head tree */
	cur = dq_hash[i];
	/* Run */
	while ( cur ) {
		/* Ignore quota mark for remove */
        if (DQ_ASP_IS_REMOVE(cur->dq.flags)) goto cur_next;

		if (DQ_ASP_CMP_TYPE(&(cur->dq), dqb)) {
			if (DQ_ASP_EQUAL(&(cur->dq), dqb)) {
				/* Hit! */
				memcpy((void *)dqb, (const void *)&cur->dq, DQ_ASP_SIZE);
				/* Release protect */
                RELEASE_PROTECT(i);
                DQ_DEBUG("end:found\n");
				return 0;
			}
		}
cur_next:
		cur = cur->next;
	}
	/* Release protect */
    RELEASE_PROTECT(i);
DQ_DEBUG("end:not found\n");
	return -1;
}
/*
 * Flush quota record from hash table to disk
 * Parameters:
 *    inout struct dq_blk *dqb	- structure for find (set veid/luid/uid/gid/flags)
 *								we flush this structure to disk
 * Return:
 *    0			- Ok
 *	  -1		- Not found	  
 *    EINVAL	- Bad input params
 */
static int dq_flush_hash(in struct dq_blk *dqb) {
	
	u32				i;
	struct ldq_blk	*cur, ldq;

DQ_DEBUG("start\n");
	// Check input params
	if (0 == dqb) return EINVAL;
	// Calc hash value
	i = dq_hash_func(dqb);
	// Set protect
    SET_PROTECT(i);
	// Set in head tree
	cur = dq_hash[i];
	// Run
	while ( cur ) {
		if (DQ_ASP_CMP_TYPE(&(cur->dq), dqb)) {
			if (DQ_ASP_EQUAL(&(cur->dq), dqb)) {
                DQ_ASP_LCOPY(&ldq,cur);
				// Release protect
                RELEASE_PROTECT(i);
				// Hit!
				dq_flush_file(&ldq);
DQ_DEBUG("end: found\n");
				return 0;
			}
		}
		cur = cur->next;
	}
	// Release protect
    RELEASE_PROTECT(i);
DQ_DEBUG("end: not found\n");
	return -1;
}
/*
 * Get quota record from hash table
 * Parameters:
 *    inout struct dq_blk *dqb	- structure for find (set veid/luid/uid/gid/flags)
 *                            we retrieve other fields
 * Return:
 *    0			- Ok
 *	  -1		- Not found	  
 *    EINVAL	- Bad input params
 */
static int dq_mark_remove_hash(in struct dq_blk *dqb) {
	
	u32				i;
	struct ldq_blk	*cur;

DQ_DEBUG("start\n");
	/* Check input params */
	if (0 == dqb) return EINVAL;
	/* Calc hash value  */
	i = dq_hash_func(dqb);
	/* Set protect */
    SET_PROTECT(i);
	/* Set in head tree */
	cur = dq_hash[i];
	/* Run */
	while ( cur ) {
		/* Ignore quota mark for remove */
		if (DQ_ASP_IS_REMOVE(cur->dq.flags)) goto cur_next;
		if (DQ_ASP_CMP_TYPE(&(cur->dq), dqb)) {
			if (DQ_ASP_EQUAL(&(cur->dq), dqb)) {
				/* Hit! */
				cur->dq.flags |= DQ_ASP_REMOVE | DQ_ASP_DIRTY;
				/* Change statistics */
				DQ_ASP_DEC_STAT(&dqstat, cur);
				/* Release protect */
                RELEASE_PROTECT(i);
                DQ_DEBUG("end: found\n");
				return 0;
			}
		}
cur_next:
		cur = cur->next;
	}
	/* Release protect */
    RELEASE_PROTECT(i);
DQ_DEBUG("end: not found\n");
	return -1;
};
/*
 * Flush all hash table to disk
 * Parameters:
 *    in u32 force		- 0/1 - If 0 flush only dirty structures; 1 - force
 * Return:
 *    None
 */
void dq_sync(in u32 force) {

	u32				i;
	struct ldq_blk	*cur, *p, *q;

DQ_DEBUG("start\n");
	for (i = 0; i < DQ_ASP_HASH_SIZE; i++) {

		// Set protect
        SET_PROTECT(i);

		cur = dq_hash[i];
        q = 0;
		while ( cur ) {
			if (force || (DQ_ASP_IS_DIRTY(cur->dq.flags))) {
                p = kmem_cache_alloc(dq_cachep, SLAB_KERNEL);
                if (NULL == p) {
                    dq_free(q);                  
                    return;
                }
                DQ_ASP_LCOPY(p,cur);
                p->next = 0;
                if ( q ) p->next = q;
                q = p;
			}
			cur = cur->next;
		}

		// Release protect
        RELEASE_PROTECT(i);

        // Flush to disk
        p = q;
        while ( p ) {
            dq_flush_file(p);
            p = p->next;
        }

        // Release memory        
        dq_free(q);
	}
DQ_DEBUG("end\n");
}
/*
 * Flush all hash table to disk and free memory
 * Parameters:
 *    in u32 veid	- veid number
 *    in int type	- type quotas for off
 * Return:
 *    0				- Ok
 *	  EINVAL		- Bad input params
 */
static int dq_off(in u32 veid, in int type) {

	u32				i, res;
	struct ldq_blk	*next, *last, *cur, *first;

DQ_DEBUG("start\n");
	if ((-1 != type) && (DQ_ASP_MAXTYPE <= type)) return EINVAL;
	// Flush to disk
    dq_sync(1);

	for (i = 0; i < DQ_ASP_HASH_SIZE; i++) {
		// Set protect
        SET_PROTECT(i);

		first = last = cur = dq_hash[i];

		while ( cur ) {
			next = cur->next;

            res = 0;
			if (0 == veid) {
				if (-1 == type) res = 1;
				else if ((cur->dq.flags & DQ_ASP_VEGUID)==(type & DQ_ASP_VEGUID)) {
					switch (type & DQ_ASP_VEGUID) {
						case DQ_ASP_VEID:
							if (0 != cur->dq.veid) res = 1;
							break;
						case DQ_ASP_UID:
						case DQ_ASP_GID:
							if (0 == cur->dq.veid) res = 1;
							break;
					}
				}
			}
			else {
				if ((cur->dq.flags & DQ_ASP_VEGUID)==(type & DQ_ASP_VEGUID)) {
					if (veid == cur->dq.veid) {
						if ((-1 == type) || (DQ_ASP_UID == type) || (DQ_ASP_GID == type)) res = 1;
					}
				}
			}
			if (0 != res) {
				// Remove quota from list
				if (last == cur) {	// Head of list
					first = last = next;
				}
				else last->next = next;
				dq_close_file(cur->fp);
				// Dec statistics
				DQ_ASP_DEC_STAT(&dqstat, cur);
				// Release memory
				kmem_cache_free(dq_cachep, (void *)cur);
			}
			else last = cur;
			cur = next;
		}
		dq_hash[i] = first;
		// Release protect
        RELEASE_PROTECT(i);
	}
DQ_DEBUG("end\n");
	return 0;
}
/*
* Calculate number of bits in flags
* Parameters:
*   in u32	flag	- Flag for testing
* Return:
*   Number of bits
*/
static u32 calc_flags(in u32 flag) {

	u32 i, cnt, mask;

	for (i = 0, cnt = 0, mask = 1; i < 31; i++, mask <<= 1) {
		if (flag & mask) cnt++;
	}
	return cnt;
}
/*
 * Open file to kernel level
 * Parameters:
 *    in const char	*fname		- Path to file open
 *    in u32		mode		- Mode for open (O_RDONLY,O_WRONLY,O_RDWR)
 *    in u32		access		- Right access
 * Return:
 *    0		- IO error
 *    !=0	- Ptr to open file
 */
static struct file *dq_open_file(
	in const char	*fname, 
	in u32			mode, 
	in u32			access) {

	char			*f;
	struct file		*fp;
	struct inode	*inode;
	int				error;

	/* Allocate 4K page for filename */
	f = getname(fname);
	error = PTR_ERR(f);
	if (IS_ERR(f)) return 0;

	/* Try open file for read */
	fp = filp_open(f, mode, access);	
	/* Release memory under filename */
	putname(f);		
	/* Check open error */
	error = PTR_ERR(fp);
	if (IS_ERR(fp)) return 0;

	/* Check presence read/write operation for file */
	if ((0 == fp->f_op->read) || (0 == fp->f_op->write)) { 
		printk("dq_open_file: null read or write ops\n");
		filp_close(fp, NULL);
		return 0; 
	}
	/* It's regular file? */
	inode = fp->f_dentry->d_inode;
	if (! S_ISREG(inode->i_mode)) {
		printk("dq_open_file: not regular file\n");
		filp_close(fp, NULL);
		return 0;
	}

	fp_cnt++;
	return fp;
}
/*
 * Close file on kernel level
 * Parameters:
 *    in struct file	*fp	- Ptr to file
 * Return:
 *    >=0		- Number of read bytes
 *    -EINVAL	- Bad input params
 */
static int dq_close_file(in struct file *fp) {

	if (NULL == fp) return -EINVAL;
	
	lock_kernel();
	filp_close(fp, NULL);
	unlock_kernel();

	if (fp_cnt > 0) fp_cnt--;
	else printk("try close, but fp_cnt=0\n");
	
	return 0;
}
/*
 * Read from file on kernel level
 * Parameters:
 *    in struct file	*fp	- Ptr to file
 *    out char	*buf		- Output buffer
 *    in u32	len			- Read len
 * Return:
 *    >=0		- Number of read bytes
 *    -EINVAL	- Bad input params
 */
/*
static int dq_read_file(in struct file *fp, out char *buf, in u32 len) {

	mm_segment_t    oldfs;
	int				ret;
		
	if ((NULL == fp) || (NULL == buf)) return -EINVAL;
	if ((NULL == fp->f_op) || (NULL == fp->f_op->read)) return -EINVAL;

	//printk("dq_read_file: len=%u; offset=%u\n", len, fp->f_pos);

	if (fp->f_mode & FMODE_READ) {
		struct inode *inode = fp->f_dentry->d_inode;
		// Set appropriate context
		oldfs = get_fs();
		set_fs(KERNEL_DS);

		ret = locks_verify_area(FLOCK_VERIFY_READ, inode, fp, fp->f_pos, len);

		if (0 == ret) {
			ret = fp->f_op->read(fp, buf, len, &fp->f_pos);
		}
		// Restore context/
		set_fs(oldfs);
	}
	else ret = -EBADF;

	return ret;
}
*/
/*
 * Write to file on kernel level
 * Parameters:
 *    in struct file	*fp		- Ptr to file 
 *    in char			*buf	- Input buffer
 *    in u32			len		- Write len
 * Return:
 *    >=0		- Number of write bytes
 *    -EINVAL	- Bad input params
 */
static int dq_write_file(in struct file *fp, in char *buf, in u32 len) {

	mm_segment_t    oldfs;
	int				ret;
		
	if ((NULL == fp) || (NULL == buf)) return -EINVAL;
	if ((NULL == fp->f_op) || (NULL == fp->f_op->write)) return -EINVAL;

	if (fp->f_mode & FMODE_WRITE) {
		//struct inode *inode = fp->f_dentry->d_inode;
		// Set appropriate context
		oldfs = get_fs();
		set_fs(KERNEL_DS);

		//ret = locks_verify_area(FLOCK_VERIFY_WRITE, inode, fp, fp->f_pos, len);

		//if (0 == ret) {
			ret = fp->f_op->write(fp, buf, len, &fp->f_pos);
		//}
		// Restore context
		set_fs(oldfs);
	}
	else ret = -EBADF;

	return ret;
}
/*
 * Write quota record to file on kernel level
 * Parameters:
 *    in struct ldq_blk *ldq
 * Return:
 *    0			- Ok
 *    EINVAL	- Bad input params
 *    EIO       - Bad write to file
 */
static int dq_flush_file(in struct ldq_blk *ldq) {

	struct file		*fp;
	int				cnt, status;

DQ_DEBUG("start\n");
	// Check input params
	if ((0 == ldq) || (0 == ldq->fp)) return EINVAL;

	fp			= ldq->fp;
	fp->f_pos	= ldq->offset;

	// Set lock. Wait under others release lock.
	if (0 != (status = dq_lock_file(fp, F_WRLCK, ldq->offset, DQ_ASP_SIZE))) {
		printk("dq_flush_file: lock error=%d\n", status);
		return status;
	}

    ldq->dq.flags &= ~DQ_ASP_DIRTY;
	cnt = dq_write_file(fp, (char *)&(ldq->dq), DQ_ASP_SIZE);

	/* Set lock. Wait under others release lock.  */
	if (0 != (status = dq_lock_file(fp, F_UNLCK, ldq->offset, DQ_ASP_SIZE))) {
		printk("dq_flush_file: unlock error=%d\n", status);
		return status;
	}

	if (cnt != DQ_ASP_SIZE) {
		printk("dq_flush_file: write error: cnt=%u, need=%u\n", cnt, DQ_ASP_SIZE);
		return EIO;
	}
DQ_DEBUG("end\n");
	return 0;
}
/*
 * Check limits for new inodes
 * Parameters:
 *    in struct dq_blk	*dqb		- Quota structure for check
 *    in u32			inodes	- Number of new inodes 
 * Return:
 *    0		 - Ok
 *    -1	 - Reject request (or hard limit exceed, or expired time for soft limit)
 *    EINVAL - Bad input params
 */
static int dq_check_inode_limit(in struct dq_blk *dqb, in u32 inodes) {

	/* Check input params */
	if (NULL == dqb) return EINVAL;
	/* Check hard limit */
	if (0 != dqb->inode_hard) {
		if ((dqb->inode_current + inodes) > dqb->inode_hard) {
			/* Reject request */
			printk("dq_check_inode_limit hard limit: try get %u inodes\n", inodes);
			return -1;
		}
	}
	/* Check soft limit */
	if (0 != dqb->inode_soft) {
		if ((dqb->inode_current + inodes) > dqb->inode_soft) {
			/* Transition soft limit */
			if (0 == dqb->cur_inode_time) {
				/* Start countdown */
				dqb->cur_inode_time = CURRENT_TIME + dqb->inode_expiration;
				printk("dq_check_inode_limit soft limit start expire: try get %u inodes\n", inodes);
			}
			/* Check countdown */
			else if (CURRENT_TIME > dqb->cur_inode_time) {
				/* Reject */
				printk("dq_check_inode_limit soft limit: time expired: try get %u inodes\n", inodes);
				return -1;
			}
		}
	}
	return 0;
}
/*
 * Check limits for new units
 * Parameters:
 *    in struct dq_blk	*dqb		- Quota structure for check
 *    in u32			units	- Number of new units
 * Return:
 *    0		 - Ok
 *    -1	 - Reject request (or hard limit exceed, or expired time for soft limit)
 *    EINVAL - Bad input params
 */
static int	dq_check_unit_limit(in struct dq_blk *dqb, in u32 units, in u32 prealloc) {

	/* Check input params */
	if (NULL == dqb) return EINVAL;
	/* Check hard limit */
	if (0 != dqb->unit_hard) {
		if ((dqb->unit_current + units) > dqb->unit_hard) {
			/* Reject */
			if (0 == prealloc)
				printk("dq_check_unit_limit hard limit: try get %u units\n", units);
			return -1;
		}
	}
	/* Check soft limit */
	if (0 != dqb->unit_soft) {
		if ((dqb->unit_current + units) > dqb->unit_soft) {
			/* Transition soft limit */
			if (0 == dqb->cur_unit_time) {
				if (0 == prealloc) {
					/* Start countdown */
					dqb->cur_unit_time = CURRENT_TIME + dqb->unit_expiration;
					printk("dq_check_unit_limit soft limit start expir: try get %u units\n", units);
				}
			}
			/* Check countdown */
			else if (CURRENT_TIME > dqb->cur_unit_time) {
				/* Reject */
				if (0 == prealloc) {
					printk("dq_check_unit_limit soft limit: time expired: try get %u units\n", units);
				}
				return -1;
			}
		}
	}
	return 0;
}
/*
 * Check for possibility allocate memory units 
 * Parameters:
 *    u32 units		- Number of memory units
 * Return:
 *    0		- Ok
 * 
 */
int dq_alloc_unit(in u32 units, in u32 prealloc) {

	struct dq_blk	dqb, dqb0;
	struct ldq_blk	*cur, *last;
	struct ldq_blk	*v, *u, *g;
	u32				i;
	int				n, vn, un, gn;
	int				status;

DQ_DEBUG1("DQ_ALLOC_UNIT=%d\n",units);
	if (0 == units) return 0;

	DQ_ASP_SET_VEGUID(&dqb0);
	/* Ignore (0, 0, 0) */
	if (0 == (dqb0.veid | dqb0.uid | dqb0.gid)) return 0;

	dqb.flags = dqb0.flags;
	v = u = g = 0;
	n = vn = un = gn = -1;	
	last = cur = 0;
	
	for (i = DQ_ASP_VEID; i < DQ_ASP_MAXTYPE; i++) {
		if ((DQ_ASP_VEID == i) && (0 == dqb0.veid)) continue;
		switch ( i ) {
			case DQ_ASP_VEID: 
				dqb.veid = dqb0.veid;
				dqb.uid = dqb.gid = 0;
				vn = dq_hash_func(&dqb);
				n = vn;
				break;
			case DQ_ASP_UID: 
				dqb.veid = dqb0.veid;
				dqb.uid = dqb0.uid;
				dqb.gid = 0;
				un = dq_hash_func(&dqb);
				if (vn == un) un = -1;
				n = un;
				break;
			case DQ_ASP_GID: 
				dqb.veid = dqb0.veid;
				dqb.uid = 0;
				dqb.gid = dqb0.gid;
				gn = dq_hash_func(&dqb); 
				if ((vn == gn) || (un == gn)) gn = -1;
				n = gn;
				break;
		}
		// Set protect
		if (-1 != n) {
			SET_PROTECT(n);
			/* Set in head tree */
			last = cur = dq_hash[n];
		}
		else cur = last;
		/* Run */
		while ( cur ) {
			/* Ignore quota mark for remove */
			if (DQ_ASP_IS_REMOVE(cur->dq.flags)) goto cur_next;
			if (cur->dq.veid != dqb.veid) goto cur_next;
			switch (cur->dq.flags & DQ_ASP_VEGUID) {
				case DQ_ASP_VEID:
					if (0 == dqb.veid) goto cur_next;
					v = cur;
					break;
				case DQ_ASP_UID:
					if ((0 == dqb.uid) || (cur->dq.uid != dqb.uid)) goto cur_next;
					u = cur;
					break;
				case DQ_ASP_GID:
					if ((0 == dqb.gid) || (cur->dq.gid != dqb.gid)) goto cur_next;
					g = cur;
					break;
				default: goto cur_next;
			}
			/* Hit! */
			if (0 != (status = dq_check_unit_limit(&(cur->dq), units, prealloc))) {
				/* If any restrictions - veid, uid, gid => reject */
				/* Release protect */
				RELEASE_PROTECT(vn);
				RELEASE_PROTECT(un);
				RELEASE_PROTECT(gn);
                DQ_DEBUG("end: reject\n");
				return status;
			}
cur_next:
			cur = cur->next;
		}
	}
	// Ok!
	if (0 != v) { 
		v->dq.unit_current += units; v->dq.flags |= DQ_ASP_DIRTY; 
	}
	if (0 != u) { 
		u->dq.unit_current += units; u->dq.flags |= DQ_ASP_DIRTY; 
	}
	if (0 != g) { 
		g->dq.unit_current += units; g->dq.flags |= DQ_ASP_DIRTY; 
	}
	// Release protect
	RELEASE_PROTECT(vn);
	RELEASE_PROTECT(un);
	RELEASE_PROTECT(gn);
DQ_DEBUG("end\n");
	return 0;
}
/*
 * Check for possibility allocate memory units without change params
 * Parameters:
 *    u32 units		- Number of memory units
 * Return:
 *    0		- Ok
 * 
 */
int dq_check_unit(in u32 units) {

	struct dq_blk	dqb, dqb0;
	struct ldq_blk	*cur, *last;
	struct ldq_blk	*v, *u, *g;
	u32				i;
	int				n, vn, un, gn;
	int				status;

DQ_DEBUG1("DQ_CHECK_UNIT=%d\n", units);
	if (0 == units) return 0;

	DQ_ASP_SET_VEGUID(&dqb0);
	// Ignore (0, 0, 0)
	if (0 == (dqb0.veid | dqb0.uid | dqb0.gid)) return 0;

	dqb.flags = dqb0.flags;
	v = u = g = 0;
	n = vn = un = gn = -1;	
	cur = last = 0;
	
	for (i = DQ_ASP_VEID; i < DQ_ASP_MAXTYPE; i++) {
		if ((DQ_ASP_VEID == i) && (0 == dqb0.veid)) continue;
		switch ( i ) {
			case DQ_ASP_VEID: 
				dqb.veid = dqb0.veid;
				dqb.uid = dqb.gid = 0;
				vn = dq_hash_func(&dqb);
				n = vn;
				break;
			case DQ_ASP_UID: 
				dqb.veid = dqb0.veid;
				dqb.uid = dqb0.uid;
				dqb.gid = 0;
				un = dq_hash_func(&dqb);
				if (vn == un) un = -1;
				n = un;
				break;
			case DQ_ASP_GID: 
				dqb.veid = dqb0.veid;
				dqb.uid = 0;
				dqb.gid = dqb0.gid;
				gn = dq_hash_func(&dqb); 
				if ((vn == gn) || (un == gn)) gn = -1;
				n = gn;
				break;
		}
		// Set protect
		if (-1 != n) {
			SET_PROTECT(n);
			// Set in head tree
			last = cur = dq_hash[n];
		}
		else cur = last;
		// Run
		while ( cur ) {
			// Ignore quota mark for remove
			if (DQ_ASP_IS_REMOVE(cur->dq.flags)) goto cur_next;
			if (cur->dq.veid != dqb.veid) goto cur_next;
			switch (cur->dq.flags & DQ_ASP_VEGUID) {
				case DQ_ASP_VEID:
					if (0 == dqb.veid) goto cur_next;
					v = cur;
					break;
				case DQ_ASP_UID:
					if ((0 == dqb.uid) || (cur->dq.uid != dqb.uid)) goto cur_next;
					u = cur;
					break;
				case DQ_ASP_GID:
					if ((0 == dqb.gid) || (cur->dq.gid != dqb.gid)) goto cur_next;
					g = cur;
					break;
				default: goto cur_next;
			}
			// Hit!
			if (0 != (status = dq_check_unit_limit(&(cur->dq), units, 0))) {
				// If any restrictions - veid, uid, gid => reject
				// Release protect
				RELEASE_PROTECT(vn);
				RELEASE_PROTECT(un);
				RELEASE_PROTECT(gn);
                DQ_DEBUG("end: reject\n");
				return status;
			}
cur_next:
			cur = cur->next;
		}
	}
	// Ok!
	// Release protect
	RELEASE_PROTECT(vn);
	RELEASE_PROTECT(un);
	RELEASE_PROTECT(gn);
DQ_DEBUG("end\n");
	return 0;
}
/*
 * Check for possibility allocate inodes 
 * Parameters:
 *    u32 inodes		- Number of memory inodes
 * Return:
 *    0		- Quota normal
 *	!=0		- Quota exceeded
 */
int dq_alloc_inode(in u32 inodes) {

	struct dq_blk	dqb, dqb0;
	struct ldq_blk	*cur, *last;
	struct ldq_blk	*v, *u, *g;
	u32				i;
	int				n, vn, un, gn;
	int				status;

DQ_DEBUG1("DQ_ALLOC_INODE\n");
	if (0 == inodes) return 0;
	DQ_ASP_SET_VEGUID(&dqb0);
	/* Ignore (0, 0, 0) */
	if (0 == (dqb0.veid | dqb0.uid | dqb0.gid)) return 0;

	dqb.flags = dqb0.flags;
	v = u = g = 0;
	n = vn = un = gn = -1;	
	cur = last = 0;
	
	for (i = DQ_ASP_VEID; i < DQ_ASP_MAXTYPE; i++) {
		if ((DQ_ASP_VEID == i) && (0 == dqb0.veid)) continue;
		switch ( i ) {
			case DQ_ASP_VEID: 
				dqb.veid = dqb0.veid;
				dqb.uid = dqb.gid = 0;
				vn = dq_hash_func(&dqb);
				n = vn;
				break;
			case DQ_ASP_UID: 
				dqb.veid = dqb0.veid;
				dqb.uid = dqb0.uid;
				dqb.gid = 0;
				un = dq_hash_func(&dqb);
				if (vn == un) un = -1;
				n = un;
				break;
			case DQ_ASP_GID: 
				dqb.veid = dqb0.veid;
				dqb.uid = 0;
				dqb.gid = dqb0.gid;
				gn = dq_hash_func(&dqb); 
				if ((vn == gn) || (un == gn)) gn = -1;
				n = gn;
				break;
		}
		/* Set protect */
		if (-1 != n) {
			SET_PROTECT(n);
			/* Set in head tree */
			last = cur = dq_hash[n];
		}
		else cur = last;
		/* Run */
		while ( cur ) {
DQ_DEBUG("P2\n");
			/* Ignore quota mark for remove */
			if (DQ_ASP_IS_REMOVE(cur->dq.flags)) goto cur_next;
			if (cur->dq.veid != dqb.veid) goto cur_next;
			switch (cur->dq.flags & DQ_ASP_VEGUID) {
				case DQ_ASP_VEID:
					if (0 == dqb.veid) goto cur_next;
					v = cur;
					break;
				case DQ_ASP_UID:
					if ((0 == dqb.uid) || (cur->dq.uid != dqb.uid)) goto cur_next;
					u = cur;
					break;
				case DQ_ASP_GID:
					if ((0 == dqb.gid) || (cur->dq.gid != dqb.gid)) goto cur_next;
					g = cur;
					break;
				default: goto cur_next;
			}
			/* Hit! */
			if (0 != (status = dq_check_inode_limit(&(cur->dq), inodes))) {
				/* If any restrictions - veid, uid, gid => reject */
				/* Release protect */
				RELEASE_PROTECT(vn);
				RELEASE_PROTECT(un);
				RELEASE_PROTECT(gn);
                DQ_DEBUG("end: reject\n");
				return status;
			}
cur_next:
			cur = cur->next;
		}
	}
	/* Ok! */
	if (0 != v) { v->dq.inode_current += inodes; v->dq.flags |= DQ_ASP_DIRTY; }
	if (0 != u) { u->dq.inode_current += inodes; u->dq.flags |= DQ_ASP_DIRTY; }
	if (0 != g) { g->dq.inode_current += inodes; g->dq.flags |= DQ_ASP_DIRTY; }
	/* Release protect */
	RELEASE_PROTECT(vn);
	RELEASE_PROTECT(un);
	RELEASE_PROTECT(gn);
DQ_DEBUG("end\n");
	return 0;
}
/*
 * Free allocated memory units 
 * Parameters:
 *    u32 units		- Number of memory units
 * Return:
 *    None
 */
void dq_free_unit(in u32 units) {

	struct dq_blk	dqb, dqb0;
	struct ldq_blk	*cur, *last;
	struct ldq_blk	*v, *u, *g;
	u32				i;
	int				n, vn, un, gn;

DQ_DEBUG1("DQ_FREE_UNIT=%d\n",units);
	if (0 == units) return;

	DQ_ASP_SET_VEGUID(&dqb0);
	// Ignore (0, 0, 0)
	if (0 == (dqb0.veid | dqb0.uid | dqb0.gid)) return;

	dqb.flags = dqb0.flags;
	v = u = g = 0;
	n = vn = un = gn = -1;	
	cur = last = 0;
	
	for (i = DQ_ASP_VEID; i < DQ_ASP_MAXTYPE; i++) {
		if ((DQ_ASP_VEID == i) && (0 == dqb0.veid)) continue;
		switch ( i ) {
			case DQ_ASP_VEID: 
				dqb.veid = dqb0.veid;
				dqb.uid = dqb.gid = 0;
				vn = dq_hash_func(&dqb);
				n = vn;
				break;
			case DQ_ASP_UID: 
				dqb.veid = dqb0.veid;
				dqb.uid = dqb0.uid;
				dqb.gid = 0;
				un = dq_hash_func(&dqb);
				if (vn == un) un = -1;
				n = un;
				break;
			case DQ_ASP_GID: 
				dqb.veid = dqb0.veid;
				dqb.uid = 0;
				dqb.gid = dqb0.gid;
				gn = dq_hash_func(&dqb); 
				if ((vn == gn) || (un == gn)) gn = -1;
				n = gn;
				break;
		}
		// Set protect
		if (-1 != n) {
			SET_PROTECT(n);
			// Set in head tree
			last = cur = dq_hash[n];
		}
		else cur = last;
		// Run
		while ( cur ) {
			// Ignore quota mark for remove
			if (DQ_ASP_IS_REMOVE(cur->dq.flags)) goto cur_next;
			if (cur->dq.veid != dqb.veid) goto cur_next;
			switch (cur->dq.flags & DQ_ASP_VEGUID) {
				case DQ_ASP_VEID:
					if (0 == dqb.veid) goto cur_next;
					v = cur;
					break;
				case DQ_ASP_UID:
					if ((0 == dqb.uid) || (cur->dq.uid != dqb.uid)) goto cur_next;
					u = cur;
					break;
				case DQ_ASP_GID:
					if ((0 == dqb.gid) || (cur->dq.gid != dqb.gid)) goto cur_next;
					g = cur;
					break;
				default: goto cur_next;
			}
cur_next:
			cur = cur->next;
		}
	}
	// Ok!
	if (0 != v) { 
		if (units >= v->dq.unit_current) v->dq.unit_current = 0;
		else v->dq.unit_current -= units; 
		if (v->dq.unit_current < v->dq.unit_soft) v->dq.cur_unit_time = (time_t)0;
		v->dq.flags |= DQ_ASP_DIRTY; 
	}
	if (0 != u) { 
		if (units >= u->dq.unit_current) u->dq.unit_current = 0;
		else u->dq.unit_current -= units; 
		if (u->dq.unit_current < u->dq.unit_soft) u->dq.cur_unit_time = (time_t)0;
		u->dq.flags |= DQ_ASP_DIRTY; 
	}
	if (0 != g) { 
		if (units >= g->dq.unit_current) g->dq.unit_current = 0;
		else g->dq.unit_current -= units; 
		if (g->dq.unit_current < g->dq.unit_soft) g->dq.cur_unit_time = (time_t)0;
		g->dq.flags |= DQ_ASP_DIRTY; 
	}
	// Release protect
	RELEASE_PROTECT(vn);
	RELEASE_PROTECT(un);
	RELEASE_PROTECT(gn);
DQ_DEBUG("end\n");
	return;
}
/*
 * Free allocated memory inodes
 * Parameters:
 *    u32 inodes		- Number of memory inodes
 * Return:
 *    None
 */
void dq_free_inode(in u32 inodes) {

	struct dq_blk	dqb, dqb0;
	struct ldq_blk	*cur, *last;
	struct ldq_blk	*v, *u, *g;
	u32				i;
	int				n, vn, un, gn;

DQ_DEBUG1("DQ_FREE_INODE\n");

	if (0 == inodes) return;
	DQ_ASP_SET_VEGUID(&dqb0);
	// Ignore (0, 0, 0)
	if (0 == (dqb0.veid | dqb0.uid | dqb0.gid)) return;

	dqb.flags = dqb0.flags;
	v = u = g = 0;
	n = vn = un = gn = -1;	
	cur = last = 0;
	
	for (i = DQ_ASP_VEID; i < DQ_ASP_MAXTYPE; i++) {
		if ((DQ_ASP_VEID == i) && (0 == dqb0.veid)) continue;
		switch ( i ) {
			case DQ_ASP_VEID: 
				dqb.veid = dqb0.veid;
				dqb.uid = dqb.gid = 0;
				vn = dq_hash_func(&dqb);
				n = vn;
				break;
			case DQ_ASP_UID: 
				dqb.veid = dqb0.veid;
				dqb.uid = dqb0.uid;
				dqb.gid = 0;
				un = dq_hash_func(&dqb);
				if (vn == un) un = -1;
				n = un;
				break;
			case DQ_ASP_GID: 
				dqb.veid = dqb0.veid;
				dqb.uid = 0;
				dqb.gid = dqb0.gid;
				gn = dq_hash_func(&dqb); 
				if ((vn == gn) || (un == gn)) gn = -1;
				n = gn;
				break;
		}
		// Set protect
		if (-1 != n) {
			SET_PROTECT(n);
			// Set in head tree
			last = cur = dq_hash[n];
		}
		else cur = last;
		// Run
		while ( cur ) {
			// Ignore quota mark for remove
			if (DQ_ASP_IS_REMOVE(cur->dq.flags)) goto cur_next;
			if (cur->dq.veid != dqb.veid) goto cur_next;
			switch (cur->dq.flags & DQ_ASP_VEGUID) {
				case DQ_ASP_VEID:
					if (0 == dqb.veid) goto cur_next;
					v = cur;
					break;
				case DQ_ASP_UID:
					if ((0 == dqb.uid) || (cur->dq.uid != dqb.uid)) goto cur_next;
					u = cur;
					break;
				case DQ_ASP_GID:
					if ((0 == dqb.gid) || (cur->dq.gid != dqb.gid)) goto cur_next;
					g = cur;
					break;
				default: goto cur_next;
			}
cur_next:
			cur = cur->next;
		}
	}
	// Ok!
	if (0 != v) { 
		if (inodes >= v->dq.inode_current) v->dq.inode_current = 0;
		else v->dq.inode_current -= inodes; 
		if (v->dq.inode_current < v->dq.inode_soft) v->dq.cur_inode_time = (time_t)0;
		v->dq.flags |= DQ_ASP_DIRTY; 
	}
	if (0 != u) { 
		if (inodes >= u->dq.inode_current) u->dq.inode_current = 0;
		else u->dq.inode_current -= inodes; 
		if (u->dq.inode_current < u->dq.inode_soft) u->dq.cur_inode_time = (time_t)0;
		u->dq.flags |= DQ_ASP_DIRTY; 
	}
	if (0 != g) { 
		if (inodes >= g->dq.inode_current) g->dq.inode_current = 0;
		else g->dq.inode_current -= inodes; 
		if (g->dq.inode_current < g->dq.inode_soft) g->dq.cur_inode_time = (time_t)0;
		g->dq.flags |= DQ_ASP_DIRTY; 
	}
	// Release protect
	RELEASE_PROTECT(vn);
	RELEASE_PROTECT(un);
	RELEASE_PROTECT(gn);
DQ_DEBUG("end\n");
	return;
}
/*
 *
 */
asmlinkage int sys_dq_asp_misc(u32 cmd, u32 a, u32 b, u32 c, u32 d) {

DQ_DEBUG("sys_dq_asp_misc(%d): cmd=%d; a=%d(%x); b=%d(%x); c=%d(%x); d=%d(%x);\n", fp_cnt, cmd, a, a, b, b, c, c, d, d);

    switch ( cmd ) {
        /* 
        in char     *a - file name; 
        out int     *b - stat data size
        */
        case DQ_ASP_CMD_GET_SD_SIZE:
            return dq_test_get_sd_size((char *)a, (int *)b);
            break;
        /* 
        in char     *a - file name; 
        out int     *b - inode number 
        */
        case DQ_ASP_CMD_GET_INODE:
            return dq_test_get_inode((char *)a, (int *)b);
            break;
        /* 
        in kdev_t   a  - device
        in int      b  - block number
        out void    *c - ptr to buffer
        in int      d  - buffer len
        */
        case DQ_ASP_CMD_GET_BLOCK:
            return dq_test_get_block((kdev_t)a, (int)b, (void *)c, (int)d);
            break;
        /* 
        in char     *a - file name; 
        out int     *b - blocksize
        */
        case DQ_ASP_CMD_GET_BLOCKSIZE:
            return dq_test_get_blocksize((char *)a, (int *)b);
            break;
        /* 
        in char     *a - file name; 
        out kdev_t   b - kdev_t
        out char    *c - mnt
        out char    *d - device name
        */
        case DQ_ASP_CMD_GET_KDEV:
            return dq_test_get_kdev((char *)a, (int *)b, (char *)c, (char *)d);
            break;
    }
    return 0;
}
/*
 *
 */
asmlinkage int sys_dq_asp(u32 cmd, u32 addr1, u32 addr2, u32 n) {

	struct dq_blk	dq;
	struct ldq_blk	ldq;
	int				status;

	//printk("sys_dq_asp(%d): cmd=%u; addr1=%x; addr2=%x; n=%u\n", fp_cnt, cmd, addr1, addr2, n);

	switch ( cmd ) {
        // Layout: addr1, addr2 - any, n - version number
        case DQ_ASP_CMD_GET_VERSION:
            put_user(DQ_ASP_VERSION, (u32 *)n);
            break;
		// Layout: if addr1 == 0, then n = type for enable, addr2 - any
		// else addr1 = ptr to dq_blk for enable, addr2 & n - any
		// Enable or quotas class, or some record
		case DQ_ASP_CMD_ENABLE:
			if (0 == addr1) {
				switch (n & DQ_ASP_VEGUID) {
					case DQ_ASP_VEID: dq_flags_type(DQ_ASP_VEID, DQ_ASP_DISABLE, DQ_ASP_OFF); break;
					case DQ_ASP_UID: dq_flags_type(DQ_ASP_UID, DQ_ASP_DISABLE, DQ_ASP_OFF); break;
					case DQ_ASP_GID: dq_flags_type(DQ_ASP_GID, DQ_ASP_DISABLE, DQ_ASP_OFF); break;
				}
				dq_sync(0);
			}
			else {
				// Transfuse user->kernel
				if (0 != copy_from_user((void *)&dq, (const void *)addr1, n)) return EIO;
				if (0 != (status = dq_flags_hash(&dq, DQ_ASP_DISABLE, DQ_ASP_OFF))) return status;
				dq_flush_hash(&dq);
			}
			break;
		// Layout: if addr1 == 0, then n = type for disable, addr2 - any
		// else addr1 = ptr to dq_blk for disable, addr2 & n - any
		// Disable or quotas class, or some record
		case DQ_ASP_CMD_DISABLE:
			if (0 == addr1) {
				switch (n & DQ_ASP_VEGUID) {
					case DQ_ASP_VEID: dq_flags_type(DQ_ASP_VEID, DQ_ASP_DISABLE, DQ_ASP_ON); break;
					case DQ_ASP_UID: dq_flags_type(DQ_ASP_UID, DQ_ASP_DISABLE, DQ_ASP_ON); break;
					case DQ_ASP_GID: dq_flags_type(DQ_ASP_GID, DQ_ASP_DISABLE, DQ_ASP_ON); break;
				}
				dq_sync(0);
			}
			else {
				// Transfuse user->kernel
				if (0 != copy_from_user((void *)&dq, (const void *)addr1, n)) return EIO;
				if (0 != (status = dq_flags_hash(&dq, DQ_ASP_DISABLE, DQ_ASP_ON))) return status;
				dq_flush_hash(&dq);
			}
			break;
		// Layout: addr1 - if addr1 == 0 - sync all table, (n!=0 - force)
		// else addr1 - ptr to dq_blk
		// addr2, n - any
		// Write hash table or one record to disk
		case DQ_ASP_CMD_SYNC:
			if (0 == addr1) dq_sync(n);
			else {
				// Transfuse user->kernel
				if (0 != copy_from_user((void *)&dq, (const void *)addr1, sizeof(dq_blk))) return EIO;
				// Try find in hash table
				dq_flush_hash(&dq);
			}
			break;
		// Layout: addr1 - ptr to dq_blk; addr2, n - any
		// Get appropriate quota record from hash table
		case DQ_ASP_CMD_GET:
			// Check input params
			if (0 == addr1) return EINVAL;
			// Transfuse user->kernel
			if (0 != copy_from_user((void *)&dq, (const void *)addr1, sizeof(dq_blk))) return EIO;
			// Try find in hash table
			if (0 != dq_get_hash(&dq)) return ESRCH;
			// Transfuse kernel->kernel
			if (0 != copy_to_user((void *)addr1, (const void *)&dq, sizeof(dq_blk))) return EIO;
			break;
		// Layout: addr1 - ptr to dq_blk; addr2 - quota file name; n - offset to quota file
		// Put quota record to hash table
		case DQ_ASP_CMD_PUT:
			// Check input params
			if ((0 == addr1) || (0 == addr2)) return EINVAL;
			// Transfuse user->kernel
			//if (0 != copy_from_user((void *)&ldq.dq, (const void *)addr1, DQ_ASP_SIZE)) return EIO;
            DQ_ASP_GET_USER_COPY((&ldq.dq),(struct dq_blk *)addr1);
			ldq.next = 0;
			ldq.offset = n;
			if (0 == (ldq.fp = dq_open_file((const char *)addr2, O_WRONLY, DQB_RIGHT))) {
					return EIO;
			}
			// Insert to hash table
            if (0 != dq_insert_hash(&ldq)) {
                dq_close_file(ldq.fp);      
                return ENOMEM;
            }
			break;
		// Layout: addr1 - ptr to dq_blk; addr2 & n - any
		// Mark to remove in hash table
		case DQ_ASP_CMD_REMOVE:
			if (0 == addr1) return EINVAL;
			// Transfuse user->kernel
			if (0 != copy_from_user((void *)&dq, (const void *)addr1, DQ_ASP_SIZE)) return EIO;
			// Mark record to remove in hash table
			if (0 != dq_mark_remove_hash(&dq)) return ESRCH;
			break;
		// Layout: addr1 - veid; addr2 - 0 or addr to dq_blk; n - type
		// Result: Flush all table to disk, free memory and reset statistics
        // If addr2 != 0 (addr1 and n - any) => addr2 ptr to structure dq_blk
        // Result: Off this quota, free memory and decrement statistics      
		case DQ_ASP_CMD_OFF:
            if (0 == addr2) return dq_off(addr1, n);
            DQ_ASP_GET_USER_COPY((&dq),(struct dq_blk *)addr2);
            return dq_off_hash(&dq);
		// Layout: addr1 - ptr to dq_stat; addr2 & n - any
		// Get quota statistics structure
		case DQ_ASP_CMD_GETSTAT:
			// Check input params
			if (0 == addr1) return EINVAL;
			// Transfuse kernel->user
			if (0 != copy_to_user((void *)addr1, (const void *)&dqstat, sizeof(struct dq_stat))) return EIO;
			break;
        // Layout: in addr1 - char *filename, out int *n - size
        // size = SD + ROUND_UP(size,8)
        // SD - Stat Data - IH_SIZE + SD_SIZE = 24 + 44 - const for each file 
        case DQ_ASP_CMD_GETSIZE:
            if ((0 == addr1) || (0 == n)) return EINVAL;
            if (0 != (status = dq_getsize((char *)addr1, (int *)n))) return status;
            break;
		// Layout: addr1, addr2, n - any
		case DQ_ASP_CMD_RESETSTAT:
			memset((void *)&dqstat, 0, sizeof(dq_stat));
			break;
	}
	return 0;
}
/*
 *  Description:
 *      For file path get stat data size
 *      Cmd: DQ_ASP_CMD_TEST_GET_SD_SIZE
 *  Parameters:
 *      in char *fpath  - file path
 *      out int *sd_size  - inode number
 *  Return:
 *      0 - Ok
 */
static int dq_test_get_sd_size(in char *fpath, out int *sd_size) {
/*    
    int             status;
    struct inode    inode_tmp;
    int             size;

    if (0 == (status = fpath2inode(fpath, &inode_tmp))) {
        size = stat_data_size(&inode_tmp);
        put_user(size, sd_size);
    }
    return status;
*/
	put_user(SD_SIZE, sd_size);
	return 0;
}
/*
 *  Description:
 *      For file path get inode number
 *      Cmd: DQ_ASP_CMD_TEST_GET_INODE
 *  Parameters:
 *      in char *fpath  - file path
 *      out int *inode  - inode number
 *  Return:
 *      0 - Ok
 */
static int dq_test_get_inode(in char *fpath, out int *inode) {
    
    int             status;
    struct inode    inode_tmp;

    if (0 == (status = fpath2inode(fpath, &inode_tmp))) {
        put_user(inode_tmp.i_ino, inode);
    }
    return status;
}
/*
 *  Description:
 *      For file path get inode number
 *      Cmd: DQ_ASP_CMD_TEST_GET_BLOCK
 *  Parameters:
 *      in kdev_t   a  - device
 *      in int      b  - block number
 *      out void    *c - ptr to buffer
 *      in int      d  - buffer len
 *  Return:
 *      0 - Ok
 */
static int dq_test_get_block(in u32 dev, in int block_number, out void *buf, in int buflen) {

    struct buffer_head  *bh;

    bh = reiserfs_bread(dev, block_number, buflen);
    if (0 == bh) return 1;
    copy_to_user(buf, bh, buflen);
    return 0;
}
/*
 *  Description:
 *      For file path get inode number
 *      Cmd: DQ_ASP_CMD_TEST_GET_BLOCKSIZE
 *  Parameters:
 *      in char *fpath      - file path
 *      out int *blocksize  - blocksize
 *  Return:
 *      0 - Ok
 */
static int dq_test_get_blocksize(in char *fpath, out int *blocksize) {

    int             status;
    struct inode    inode_tmp;

    if (0 == (status = fpath2inode(fpath, &inode_tmp))) {
        put_user(inode_tmp.i_blksize, blocksize);
    }
    return status;
}
/*
 *  Description:
 *      For file path get inode number
 *      Cmd: DQ_ASP_CMD_TEST_GET_KDEV
 *  Parameters:
 *      in char *fpath - file path
 *      out int *kdev  - kdev
 *      out char *mnt  - mount point for this device
 *      out char *dev_name - device name for this path
 *  Return:
 *      0       - Ok
 *      ENOMEM  - Can't allocate memory
 *      ESRCH   - Can't find mount point (???)
 *      EIO     - Can't copy to user address space
 */
static int dq_test_get_kdev(in char *fpath, out int *kdev, out char *mnt, out char *dev_name) {

    int                 status;
    struct nameidata    nd;
    char                *s, *buffer;
    struct dentry       *de;

    if (0 == (status = user_path_walk_link(fpath, &nd))) {
        put_user(nd.dentry->d_inode->i_dev, kdev);
        buffer = kmalloc(4096, GFP_ATOMIC);
        if (0 == buffer) return ENOMEM;
        s = d_path(de = nd.mnt->mnt_root, nd.mnt, buffer, 4096);
        if (0 == s) { kfree(buffer); return ESRCH; }
        if (0 != copy_to_user((void *)mnt, (const void *)s, strlen(s))) {
            status = EIO;
        }
        else {
            strcpy(buffer, nd.mnt->mnt_devname?nd.mnt->mnt_devname:"none");
            if (0 != copy_to_user((void *)dev_name, (const void *)buffer, strlen(buffer))) {
                status = EIO;
            }
        }
        kfree(buffer);
    }
    return status;
}
/*
 *  Description:
 *      For file path get inode structure
 *  Parameters:
 *      in char *fpath          - file path
 *      out struct inode *inode - output inode structure
 *  Return:
 *      0 - Ok
 *      1 - Error
 */
static int fpath2inode(in char *fpath, out struct inode *inode) {

    struct nameidata    nd;
    int                 err;

    if (0 == (err = user_path_walk_link(fpath, &nd))) memcpy(inode, nd.dentry->d_inode, sizeof(struct inode));
    return err;
    /*
	char		                *tmp;
	struct file                 *f;
    int                         error;

	tmp = getname(fpath);
	error = PTR_ERR(tmp);
	if (IS_ERR(tmp)) {
		printk("%s error: getname\n", __FUNCTION__);
		return 1;
	}

    f = filp_open(tmp, O_RDONLY, 0);
	putname(tmp);

	error = PTR_ERR(f);
	if (IS_ERR(f)) {
		printk("fpath2inode error: filp_open\n");
		return 1;
	}

    memcpy((void *)inode, (void *)f->f_dentry->d_inode, sizeof(struct inode));

    filp_close(f, NULL);
    return 0;
    */
}
#ifdef DQ_TEST
static int try_key(struct inode *inode, 
                   struct path *path,
                    int offset, 
                    int type, 
                    struct item_head **ih, 
                    struct buffer_head **bh) {

    char            s[64];
    struct cpu_key  key;
    int             retval;

    get_type(s, type);
    make_cpu_key (&key, inode, offset, type, 3);
    pr_cpu_key("cpu_key:", &key);
    pr_inode(inode);

    retval = search_item(inode->i_sb, &key, path);
    switch ( retval ) {
        case IO_ERROR: 
            printk("%s search error: IO_ERROR\n", s); 
            return -1;
        case ITEM_NOT_FOUND: 
            printk("%s search error: ITEM_NOT_FOUND (try_key)\n", s); 

            look_for(inode, 1, TYPE_DIRECT);
            look_for(inode, 2, TYPE_DIRECT);

            look_for(inode, 1, TYPE_INDIRECT);
            look_for(inode, 2, TYPE_INDIRECT);

            look_for(inode, 1, TYPE_ANY);
            look_for(inode, 2, TYPE_ANY);
            return -1;
        case ITEM_FOUND: 
            DQ_DEBUG("%s search Ok\n", s);
            *bh = get_bh(path);
	        *ih = get_ih(path);
            break;
    }

    return 0;
}

static void pr_inode(struct inode *inode) {

	printk("i_ino       =%d\n", (int)inode->i_ino);
	printk("i_uid       =%d\n", (int)inode->i_uid);
	printk("i_gid       =%d\n", (int)inode->i_gid);
	printk("i_size      =%ld\n", (long)inode->i_size);
	printk("i_blksize   =%d\n", (int)inode->i_blksize);
	printk("i_blocks    =%d\n", (int)inode->i_blocks);
	printk("i_flags     =0x%x\n", (int)inode->i_flags);
	printk("i_attr_flags=0x%x\n", (int)inode->i_attr_flags);
	printk("i_generation=%d\n", (int)inode->i_generation);
	printk("i_state     =%d\n", (int)inode->i_state);
	printk("sb_blocksize=%d\n", (int)inode->i_sb->s_blocksize);
}

static void pr_dentry(struct dentry *de) {

    printk("name=%s\n", de->d_name.name);
    printk("flags=%d\n", de->d_flags);
}

static void pr_stat_data(struct stat_data *st) {

    printk("sd_mode=%d\n",  st->sd_mode);
    printk("sd_size=%ld\n", (long)st->sd_size);
    printk("sd_uid=%d\n",  st->sd_uid);
    printk("sd_gid=%d\n",  st->sd_gid);
    printk("sd_blocks=%d\n",  st->sd_blocks);
}


static void pr_de_head(struct reiserfs_de_head *deh) {
/*
  __u32 deh_offset;		// third component of the directory entry key
  __u32 deh_dir_id;		// objectid of the parent directory of the object, that is referenced
					   by directory entry
  __u32 deh_objectid;	// objectid of the object, that is referenced by directory entry
  __u16 deh_location;	// offset of name in the whole item
  __u16 deh_state;		// whether 1) entry contains stat data (for future), and 2) whether
					   entry is hidden (unlinked)
*/

    printk("deh_offset  =%d\n", deh->deh_offset);
    printk("deh_dir_id  =%d\n", deh->deh_dir_id);
    printk("deh_objectid=%d\n", deh->deh_objectid);
    printk("deh_location=%d\n", deh->deh_location);
    printk("deh_state   =%d\n", deh->deh_state);
}


static void pr_dir_entry(struct reiserfs_dir_entry *de) {

    char name[256];

/*
    int de_entry_num;
    struct reiserfs_de_head * de_deh;
    int de_entrylen;
    int de_namelen;
    char * de_name;
    char * de_gen_number_bit_string;
    __u32 de_dir_id;
    __u32 de_objectid;
    struct cpu_key de_entry_key;
*/
    printk("de_item_num =%d\n", de->de_item_num);
    printk("de_entry_num=%d\n", de->de_entry_num);
    printk("de_dir_id   =%d\n", de->de_dir_id);
    printk("de_objectid =%d\n", de->de_objectid);
    printk("de_namelen  =%d\n", de->de_namelen);
    memset(name, 0, sizeof(name));
    strncpy(name, de->de_name, de->de_namelen);
    printk("de_name     =%s\n", name);
    reiserfs_warning("de_bh=%b\n", de->de_bh);
    reiserfs_warning("de_ih=%h\n", de->de_ih);
    reiserfs_warning("de_entry_key=%K\n", de->de_entry_key);
    pr_de_head(de->de_deh);
}
/*
static void pr_essential_constants(void) {

    printk("BLKH_SIZE       = %d\n", BLKH_SIZE);
    printk("KEY_SIZE        = %d\n", KEY_SIZE);       // sizeof(struct key)
    printk("IH_SIZE         = %d\n", IH_SIZE);        // sizeof(struct item_head)
    printk("MIN_ITEM_LEN    = %d\n", MIN_ITEM_LEN);
    printk("MAX_ITEM_LEN    = %d\n", MAX_ITEM_LEN);
    printk("SD_SIZE         = %d\n", SD_SIZE);        // sizeof(struct stat_data)
    printk("DEH_SIZE        = %d\n", DEH_SIZE);       // sizeof(struct reiserfs_de_head)
    printk("EMPTY_DIR_SIZE  = %d\n", EMPTY_DIR_SIZE);
    printk("EMPTY_DIR_SIZE_V1= %d\n", EMPTY_DIR_SIZE_V1);
    printk("DC_SIZE         = %d\n", DC_SIZE);        // sizeof(struct disk_child)
    printk("UNFM_P_SIZE     = %d\n", UNFM_P_SIZE);
}
*/
#endif
/*
 * Get file size
 * SD - Stat Data - IH_SIZE + SD_SIZE = 24 + 44 - const for each file 
 * DIR - IH_SIZE + SD_SIZE + ROUND_UP(size,8)
 * FILE (direct) - IH_SIZE + SD_SIZE + ROUND_UP(size,8)
 * FILE (indirect) - IH_SIZE + SD_SIZE + 4*N_INDIR_ITEMS + BLK_SIZE*N_INDIR_ITEMS
 * Attention: Work only for DIR & REG!
 * Parameters:
 *      in char *fpath  - file/dir path
 *      out int *n      - file/dir size
 * Return:
 *      0       - Ok
 */
static int dq_getsize(in char *fpath, out int *n) {

    struct inode            inode;
    int                     status, sd_size;

    if (0 != (status = fpath2inode(fpath, &inode))) return status;
    sd_size = IH_SIZE + SD_SIZE + ROUND_UP(inode.i_size);
    put_user(sd_size, n);
    return 0;
}
/*
 *
 */
#ifdef DQ_TEST
static void pr_cpu_key(char *s, struct cpu_key  *key) {

    reiserfs_warning("%s key=%K\n", s, key);
}

static void pr_ih(char *s, struct item_head *ih) {

    reiserfs_warning("%s ih=%h\n", s, ih);
}

static void pr_bufh(char *s, struct buffer_head *bufh) {

    reiserfs_warning("%s bufh=%z\n\n", s, bufh);
}

static void pr_blockh(char *s, struct block_head *blockh) {

    reiserfs_warning("%s blockh=%b\n", s, blockh);
}

static int look_for(struct inode *inode, int offset, int type) {

    char                s[16], s1[64];
    struct cpu_key      key;
    int                 ret;
    struct item_head    *ih; 
    struct buffer_head  *bh;


    INITIALIZE_PATH(path);

    get_type(s, type);
    make_cpu_key (&key, inode, offset, type, 3);

    pr_cpu_key("cpu_key:", &key);

    ret = search_item(inode->i_sb, &key, &path);
    switch ( ret ) {
        case IO_ERROR: 
            printk("%s:IO_ERROR\n",s); 
            return ret;
        case ITEM_NOT_FOUND:
            sprintf(s1,"%s:ITEM_NOT_FOUND", s);
            printk("%s\n",s1); 
            break;
        case ITEM_FOUND: 
            sprintf(s1,"%s:ITEM_FOUND", s);
            printk("%s\n",s1);
            break;
        default:
            printk("%s:UNKNOWN RESULT\n", s);
            return ret;
    }

	ih = get_ih(&path);
    pr_ih(s1, ih);
    bh = get_bh(&path);
    pr_blockh(s1, bh);

    if (is_direct_le_ih(ih)) printk("DIRECT IH\n");
    else if (is_indirect_le_ih(ih)) printk("INDIRECT IH\n");
    else if (is_direntry_le_ih(ih)) printk("DIRENTRY IH\n");
    else if (is_statdata_le_ih(ih)) {
        printk("STAT_DATA IH\n");
        pr_stat_data(B_I_STAT_DATA(bh,ih));
    }
    else printk("??? IH\n");

    pathrelse(&path);
    return ret;
}

static void look_for_ok(struct inode *inode, int offset, int type, int cnt) {

    char                s[16];
    struct cpu_key      key;
    int                 i, ret;

    get_type(s, type);
    for (i = 0; i < cnt; i++) {
        INITIALIZE_PATH(path);
        make_cpu_key (&key, inode, offset+i, type, 3);
        ret = search_item(inode->i_sb, &key, &path);
        if (ITEM_FOUND == ret) {
            printk("%s:OK!:offset=%d;i=%d\n", s, offset, i);
        }
        pathrelse(&path);
    }
    return;
}

static void get_type(char *s, int type) {

    switch ( type ) {
        case TYPE_STAT_DATA:    strcpy(s, "SD"); break;
        case TYPE_INDIRECT:     strcpy(s, "INDIR"); break;
        case TYPE_DIRECT:       strcpy(s, "DIRECT"); break;
        case TYPE_DIRENTRY:     strcpy(s, "DIRENTRY"); break;
        case TYPE_ANY:          strcpy(s, "ANY"); break;
        default:                strcpy(s, "UNKNOWN"); break;
    }
}

#endif
/*
 * Release tree
 */
static void dq_free(struct ldq_blk *p) {
    
    struct ldq_blk *next;

    while ( p ) {
        next = p->next;
    	// Release memory
		kmem_cache_free(dq_cachep, (void *)p);
        p = next;
    }
}
