/*
*
* Author:              Vadim V. Sinitsyn
* E-mail:              sinvv@asplinux.ru
*
* $Header: /home/cvs/ASPcomplete/utils/dqlib.c,v 1.19 2000/10/17 17:08:27 sinvv Exp $
*
* Remarks:
*
* Last correct: 17/Oct/00
*
*/
#include <string.h>
#include <sys/types.h>
#include <dirent.h>
#include <errno.h>
#include <time.h>
#include <linux/kdev_t.h>

#include <asm/unistd.h>
#include <sys/vfs.h>
#include <linux/unistd.h>
#include "dqlib.h"

// Syscall
_syscall3(int, env_create, unsigned, envid, unsigned, flags, unsigned, addr);
_syscall4(int, dq_asp, unsigned, cmd, unsigned, addr1, unsigned, addr2, unsigned, n);
_syscall5(int, dq_asp_misc, unsigned, cmd, unsigned, a, unsigned, b, unsigned, c, unsigned, d);

/*
    Next defines from: linux/reiserfs_fs.h
    Check after change
*/
/* n must be power of 2 */
#define _ROUND_UP(x,n) (((x)+(n)-1u) & ~((n)-1u))
// to be ok for alpha and others we have to align structures to 8 byte
// boundary.
// FIXME: do not change 4 by anything else: there is code which relies on that
#define ROUND_UP(x) _ROUND_UP(x,8LL)
#define IH_SIZE         24
#define SD_SIZE         44
#define DEH_SIZE        24

#undef DQLIB_DEBUG
#ifdef DQLIB_DEBUG
static void print_lstat(in struct stat *st);
static int print_statfs(in char *fpath);
static int  get_blksize(in char *fpath, out int *blksize);
static int  get_sd_size(in char *fpath, out int *sd_size);
#endif

static int	is_reiserfs(in char *fpath, out int *flag_reiserfs);
static int conv_path(char *dst, char *fpath);
static char *gostr(char *s);
static int str_istime(char *s, int n);
/*
 * This line is empty or comments?
 * Parameters:
 *   in char *buf	- String for checking
 * Return:
 *    0		- Empty (only space, tab, cr, lf)
 *    -1	- Not empty
 */
int is_emptyline(in char *buf) {

	char	c;

	while (0 != (c = *buf++)) {
		if (isspace(c)) continue;
		if (('#' == c) || (';' == c)) return 0; /* Comments */
		return -1;
	}
	return 0;
}
/*
 * Print header for structure dqblk
 * Parameters:
 *   None
 * Return:
 *	 None
 */
void print_dqblk_header(void) {

	dprint("%5s%4s%4s%3s%9s%9s%9s%18s%7s%6s%6s%6s%18s%7s\n",
		"veid", "uid", "gid", "fl",
		"ucur", "usoft", "uhard", "ucur", "uexp",
		"icur", "isoft", "ihard", "icur", "iexp");
}
/*
 * Print structure dqblk
 * Parameters:
 *   in struct dq_blk *dq	- Ptr to printing structure
 *   in __u32 format			- Output format
 *   in __u32 unit_out_format - Ouput format for unit (Bytes/Kb/Mb)
 * Return:
 *	None
 */
static char *monthes[12] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };

void print_dqblk(in struct dq_blk *dq, in __u32 format, in __u32 unit_out_format) {

	__u32	    u_current, u_soft, u_hard;
    char        flag[32], ut[256], ue[256], it[256], ie[256];
    char        uc[256], us[256], uh[256];
    struct tm   *tm;

    // Set units
	u_current	= dq->unit_current;
	u_soft		= dq->unit_soft;
	u_hard		= dq->unit_hard;

	if (DQ_ASP_FORMAT_UNIT_KB == unit_out_format) {
		u_current = (u_current + 512)/KB;
		u_soft = (u_soft + 512)/KB;
		u_hard = (u_hard + 512)/KB;
        sprintf(uc, "%uK", u_current);
        sprintf(us, "%uK", u_soft);
        sprintf(uh, "%uK", u_hard);
	}
	else if (DQ_ASP_FORMAT_UNIT_MB == unit_out_format) {
		u_current = (u_current + (512*KB))/MB;
		u_soft = (u_soft + (512*KB))/MB;
		u_hard = (u_hard + (512*KB))/MB;
        sprintf(uc, "%uM", u_current);
        sprintf(us, "%uM", u_soft);
        sprintf(uh, "%uM", u_hard);
	}
    else {
        dq_unit2str(u_current, uc, 1);
        dq_unit2str(u_soft, us, 1);
        dq_unit2str(u_hard, uh, 1);
    }

	// Set flags
	flag[0] = flag[1] = ' '; flag[2] = 0;
	if (DQ_ASP_IS_DISABLE(dq->flags)) flag[0] = 'D';
	switch (dq->flags & DQ_ASP_VEGUID) {
		case DQ_ASP_VEID:   flag[1] = 'V'; break;
		case DQ_ASP_UID:    flag[1] = 'U'; break;
		case DQ_ASP_GID:    flag[1] = 'G'; break;
		default:            flag[1] = '@'; break;
	}

    // Set time
    //dq_time2str(dq->cur_unit_time, ut, 1);
    if (0 != dq->cur_unit_time) {
        tm = localtime((time_t *)&dq->cur_unit_time);
        sprintf(ut, "%3s %02u %02u:%02u %04u", monthes[tm->tm_mon], tm->tm_mday, tm->tm_hour, tm->tm_min, 1900+tm->tm_year);
    } 
    else strcpy(ut, "0");
    dq_time2str(dq->unit_expiration, ue, 1);
    //dq_time2str(dq->cur_inode_time, it, 1);
    if (0 != dq->cur_inode_time) {
        tm = localtime((time_t *)&dq->cur_inode_time);
        sprintf(it, "%3s %02u %02u:%02u %04u", monthes[tm->tm_mon], tm->tm_mday, tm->tm_hour, tm->tm_min, 1900+tm->tm_year);
    }
    else strcpy(it, "0");
    dq_time2str(dq->inode_expiration, ie, 1);

	if (DQ_ASP_FORMAT_COLUMN == format) {
		dprint("\tveid: %u\n", dq->veid);
		dprint("\tuid: %u\n", dq->uid);
		dprint("\tgid: %u\n", dq->gid);
        dprint("\tflags: %s (%x)\n", flag, dq->flags);

        dprint("\tunit_current: %s\n", uc);
		dprint("\tunit_soft: %s\n", us);
		dprint("\tunit_hard: %s\n", uh);

        dprint("\tcurrent_unit_time: %s\n", ut);
        dprint("\tunit_expiration: %s\n", ue);

		dprint("\tinode_current: %u\n", dq->inode_current);
		dprint("\tinode_soft: %u\n", dq->inode_soft);
		dprint("\tinode_hard: %u\n", dq->inode_hard);

        dprint("\tcurrent_inode_time: %s\n", it);
        dprint("\tinode_expiration: %s\n", ie);

	}
	else if (DQ_ASP_FORMAT_LINE == format) {
        dprint("%5u%4u%4u%3s%9s%9s%9s%18s%7s%6u%6u%6u%18s%7s\n",
		dq->veid, dq->uid, dq->gid, flag,
		uc, us, uh, ut, ue, 
		dq->inode_current, dq->inode_soft, dq->inode_hard, 
        it, ie);
	}
}
/*
 * Debug print for parse options
 * Parameters:
 *   in char ret - Return code from last getopt
 * Return:
 *    None
 */
void print_getopt_result(in char ret) {

	dprint("ret=%d (%c)\n", ret, ret);
	dprint("optind=%d\n", optind);
	dprint("opterr=%d\n", opterr);
	dprint("optopt=%d (%c)\n", optopt, optopt);
	dprint("optarg=%s\n", optarg);
}
 /*
 * Check file presence
 * Parameters:
 *    in const char	*path	- Path to checking file
 *    in int		uid		- Owner
 *    in __u32		access	- Access rights
 * Return:
 *    0 			- Ok
 *    ESRCH 		- File not found (see errno for detail)
 *    EACCES		- Bad owner or access rights
 */
int file_exist(
	in const char		*fpath,
    in int				uid,
    in __u32				access) {
 	
	struct stat buf;
	int			ret;
		
	ret = stat(fpath, &buf); 
	if (0 == ret) {
		/* Check file: access and owner */
		if (uid != buf.st_uid) {
			dprint("Bad uid=%x\n", buf.st_uid);
			return EACCES;
		}
		if ((access == (buf.st_mode & access)) &&
			(0 == (buf.st_mode & ~access))) return 0;
		else {
			dprint("Bad access= %o\n", buf.st_mode);
			return EACCES;
		}
	}
	dprint("Not find file %s\n", fpath);
	return ESRCH;
}
/*
 * Check magic number for quota data base file
 * Parameters: 
 *   in const char *dqb_path - Path to dqb file
 * Return:
 *    0			- Ok
 *    ESRCH		- File not found (see errno for more detail)
 *    EACCES	- Bad rights
 *    EIO		- Can't open/read or close
 */
int check_magic(in const char *dqb_path) {

	int 			ret;
	int				fd;
	char 			buf[32] = { 0 };
	struct flock	lock;

	/* exist */
	if (0 != (ret = file_exist(dqb_path, ROOT_UID, DQB_RIGHT))) return ret;
	/* open */
	dq_open(fd, O_RDONLY, dqb_path);
	/* wait unlock */
	dq_readlock(fd,&lock,0,MAGICLEN);
	/* read */
	dq_read(fd, buf, MAGICLEN, dqb_path);
	/* close */
	dq_close(fd, dqb_path);
	/* check */
	if (0 != strncasecmp(buf, MAGIC, MAGICLEN)) {
		dprint("Bad magic string: %s\n", buf);
		return EBADF;
	}
	return 0;
}
/*
 * Read quota data base counter
 * Parameters: 
 *   in const char *dqb_path	- Path to dqb file
 *   out __u32		*cnt		- Counter value
 * Return:
 *    0			- Ok
 *    ESRCH		- File not found (see errno for more detail)
 *    EACCES	- Bad rights
 *    EIO		- Can't open/read/seek or close
 */
int read_dqb_cnt(in const char *dqb_path, out __u32 *cnt) {

	int				ret;
	int				fd;
	struct flock	lock;

	/* Check magic number for dqb */
	if (0 != (ret = check_magic(dqb_path))) return ret;
	/* open */
	dq_open(fd, O_RDONLY, dqb_path);
	/* wait unlock */
	dq_readlock(fd,&lock,MAGICLEN+sizeof(__u32),sizeof(__u32));
	/* position to counter */
	dq_seek(fd, (MAGICLEN+sizeof(__u32)), dqb_path);
	/* read counter */	
	dq_read(fd, cnt,sizeof(__u32),dqb_path);
	/* close */
	dq_close(fd, dqb_path);
	return 0;
}
/*
 * Read Quota Data Base version
 * Parameters: 
 *   in const char *dqb_path	- Path to dqb file
 *   out __u32		*cnt		- Counter value
 * Return:
 *    0			- Ok
 *    ESRCH		- File not found (see errno for more detail)
 *    EACCES	- Bad rights
 *    EIO		- Can't open/read/seek or close
 */
int read_dqb_version(in const char *dqb_path, out __u32 *version) {

	int				ret;
	int				fd;
	struct flock	lock;

	/* Check magic number for dqb */
	if (0 != (ret = check_magic(dqb_path))) return ret;
	/* open */
	dq_open(fd, O_RDONLY, dqb_path);
	/* wait unlock */
	dq_readlock(fd,&lock,MAGICLEN,sizeof(__u32));
	/* position to counter */
	dq_seek(fd, MAGICLEN, dqb_path);
	/* read counter */	
	dq_read(fd, version, sizeof(__u32), dqb_path);
	/* close */
	dq_close(fd, dqb_path);
	return 0;
}
/*
 * Write quota data base counter
 * Parameters: 
 *   in const char	*dqb_path	- Path to dqb file
 *   in __u32			cnt			- Counter value
 * Return:
 *    0			- Ok
 *    ESRCH		- File not found (see errno for more detail)
 *    EACCES	- Bad rights
 *    EIO		- Can't open/write/seek or close
 */
int write_dqb_cnt(in const char *dqb_path, in __u32 cnt) {

	int				ret;
	int				fd;
	struct flock	lock;

	/* Check magic number for dqb */
	if (0 != (ret = check_magic(dqb_path))) return ret;
	/* open */
	dq_open(fd, O_WRONLY, dqb_path);
	/* wait unlock */
	dq_writelock(fd,&lock,MAGICLEN+sizeof(__u32),sizeof(__u32));
	/* position to counter */
	dq_seek(fd, (MAGICLEN+sizeof(__u32)), dqb_path);
	/* read counter */	
	dq_write(fd, &cnt, sizeof(__u32), dqb_path);
	/* close */
	dq_close(fd, dqb_path);
	return 0;
}
/*
 * Create empty quota data base file (Magic number and zero counter)
 * Parameters: 
 *   in const char	*dqb_path	- Path to dqb file
 * Return:
 *    0			- Ok
 *    ESRCH		- File not found (see errno for more detail)
 *    EACCES	- Bad rights
 *    EIO		- Can't create (with 0600 permissions)/bad read or close
 */
int create_empty_dqb(in const char *dqb_path) {

	int			fd;
	char		*buf = MAGIC;
	__u32			ver = DQB_VERSION, cnt = 0;

	/* check root policy */
	check_root;	
	/* create empty file */
	dq_create(fd, dqb_path);
	/* write magic */
	dq_write(fd, buf, MAGICLEN, dqb_path);
	/* write version number */
	dq_write(fd, &ver, sizeof(__u32), dqb_path);
	/* write zero counter */
	dq_write(fd, &cnt, sizeof(__u32), dqb_path);
	/* close */
	dq_close(fd, dqb_path);
	return 0;
}
/*
 * Validate structure dqblk
 * Parameters:
 *   in struct dq_blk *dq - Ptr to validate structure
 * Return:
 *   0		- Ok
 *   -1		- Not valide
 */
int validate_dqblk(in struct dq_blk *dq) {

	if (dq->unit_soft > dq->unit_hard) return -1;
	if ((0 == dq->unit_soft) || (0 == dq->unit_hard)) return -1;
	if (dq->inode_soft > dq->inode_hard) return -1;
	if ((0 == dq->inode_soft) || (0 == dq->inode_hard)) return -1;
	return 0;
}
/*
 * Free list
 * Parameters:
 *   in struct ldq_blk *dq - ptr to list structures
 * Return:
 *   None
 */
void free_list(in struct ldq_blk *dq) {

	struct ldq_blk *p, *q;

	q = dq;
	while ( q ) {
		p = q->next;
		dq_free(q);
		q = p;
	}
}
/*
 * Print list
 * Parameters:
 *   in struct ldq_blk *dq - ptr to list structures
 * Return:
 *   None
 */
void print_list(in struct ldq_blk *dq) {

	struct ldq_blk *p, *q;

	q = dq;
	while ( q ) {
		p = q->next;
		dprint("next=%p\n", p);
		print_dqblk(&q->dq, DQ_ASP_FORMAT_COLUMN, 0);
		q = p;
	}
}

/*
* Calculate number of bits in flags
* Parameters:
*   __u32	flag	- Flag for testing
* Return:
*   Number of bits
*/
__u32 calc_flags(__u32 flag) {

	__u32 i, cnt, mask;

	for (i = 0, cnt = 0, mask = 1; i < 31; i++, mask <<= 1) {
		if (flag & mask) cnt++;
	}
	return cnt;
}
/*
 * Check which VE run now
 * 0 - main VE (VE 0)
 * -1 - VE system not run now
 * >0 - appropriate VE (slave)
 * Parameters:
 *    None
 * Return:
 *    0		- Ok (If VE=0 or not run)
 *    -1	- slave VE
 */
int is_ve0(void) {
	
	int		ve;

	ve = my_veid;
	if ((VEID_0 == ve) || (NOTRUN_VE == ve)) return 0;
	return -1;
}
/*
 * Read dqb to memory
 * Parameters:
 *    in const char *dqb_path	- Path to dqb file
 *    out struct dq_blk **dq		- Ptr to output table
 *    out __u32 *cnt				- Counter
 *    out __u32 *ver				- Version
 * Return:
 *    0				- Ok
 *    EINVAL		- NULL ptr to dqb
 *    ESRCH			- File not found (see errno for more detail)
 *    EACCES		- Bad rights
 *    EIO			- Can't open/read/seek or close
 */
int read_dqb(in const char *dqb_path, out struct dq_blk **dq, out __u32 *cnt, out __u32 *ver) {

	int					status;
	int					fd;
	struct dq_blk		*q;
	struct flock		lock;

	/* Check input params */
	if ((NULL == dqb_path) || (NULL == dq) || (NULL == cnt)) return EINVAL;

	if (0 != (status = read_dqb_version(dqb_path, ver))) return status;
	if (0 != (status = read_dqb_cnt(dqb_path, cnt))) return status;
	/* Alloc for table */
	dq_malloc(q, DQ_ASP_ROUND_UP(*cnt * DQ_ASP_SIZE));
	/* open */
	dq_open(fd, O_RDONLY, dqb_path);
	/* wait unlock */
	dq_readlock(fd,&lock,MAGICLEN+2*sizeof(__u32),*cnt*DQ_ASP_SIZE);
	/* position to begin table */
	dq_seek(fd, (MAGICLEN+2*sizeof(__u32)), dqb_path);
	/* read table */	
	dq_read(fd, q, *cnt*DQ_ASP_SIZE, dqb_path);
	/* close */
	dq_close(fd, dqb_path);
	*dq = q;
	return 0;		
}
/*
 * Find quota to quotas table
 * Parameters:
 *    in struct dq_blk	*dq	- Ptr to quotas table
 *    in __u32			cnt	- Len quotas table
 *    in struct dq_blk	*q	- Ptr to quota
 *    out struct dq_blk	**r	- Ptr to result quota
 * Return:
 *    0			- Succesfull lookup
 *    -1		- Can't find
 *    EFAULT	- Panic!
 */
int find_quotas_table(
	in struct dq_blk	*dq,
	in __u32			cnt,
	in struct dq_blk	*q,
	out struct dq_blk	**r) {

	__u32				i;
	struct dq_blk	*p;
	
	for (i = 0, p = dq; i < cnt; i++, p++) {
		if (0 == same_quotas(q, p)) {
			*r = p;
			return 0;
		}
	}
	return -1;
}
/*
 * At the same quotas?
 * Parameters:
 *    in struct dq_blk	*q	- 1st quota for comparing
 *    in struct dq_blk	*p  - 2nd quota for comparing
 * Return:
 *    0			- Succesfull lookup
 *    -1		- Any types
 *    EFAULT	- Panic!
 */
int same_quotas(in struct dq_blk *q, in struct dq_blk *p) {

    // Check input params
	if ((NULL == q) || (NULL == p)) return EINVAL;
    // At the same type?
	if (0 == DQ_ASP_CMP_TYPE(q,p)) return EFAULT;

	switch (p->flags & DQ_ASP_VEGUID) {
		case DQ_ASP_VEID:
			if (q->veid != p->veid) return -1;
			break;
		case DQ_ASP_UID:
			if ((q->uid != p->uid) || (q->veid != p->veid)) return -1;
			break;
		case DQ_ASP_GID:
			if ((q->gid != p->gid) || (q->veid != p->veid)) return -1;
			break;
	}
	return 0;
}
/*
 * Set quota
 * Parameters:
 *    out struct dq_blk	*dest	- Destination quota
 *    in struct dq_blk	*src	- Source quota
 * Return:
 *    0			- Ok
 *    EINVAL	- NULL input ptrs
 */
int quota_set(out struct dq_blk *dest, in struct dq_blk *src) {

    // Check input params
	if ((NULL == dest) || (NULL == src)) return EINVAL;

	dest->unit_soft			= src->unit_soft;
	dest->unit_hard			= src->unit_hard;
	if (0 == src->unit_expiration) dest->unit_expiration = DEFAULT_EXPIRATION;
	else dest->unit_expiration	= src->unit_expiration;
	dest->cur_unit_time		= src->cur_unit_time;
	dest->inode_soft		= src->inode_soft;
	dest->inode_hard		= src->inode_hard;
	if (0 == src->inode_expiration) dest->inode_expiration = DEFAULT_EXPIRATION;
	else dest->inode_expiration	= src->inode_expiration;
	dest->cur_inode_time		= src->cur_inode_time;
	return 0;
}
/*
 * Flush quotas table to dqb
 * Parameters:
 *    in const char		*dqb_path	- Path to dqb file
 *    in struct dq_blk	*dq			- Ptr to quotas table
 *    in struct dq_blk	*p			- 2nd quota for comparing
 * Return:
 *    0				- Succesfull lookup
 *    EINVAL		- NULL ptr to dqb
 *    ESRCH			- File not found (see errno for more detail)
 *    EACCES		- Bad rights
 *    EIO			- Can't open/write/seek or close
 */
int flush_tbl2dqb(in const char *dqb_path, in struct dq_blk *dq, in __u32 cnt) {

	int				status;
	__u32			ver;
	int				fd;
	struct flock	lock;

	// Check input params
	if ((NULL == dqb_path) || (NULL == dq)) return EINVAL;
	// For verification
	if (0 != (status = read_dqb_version(dqb_path, &ver))) return status;	
	// open
	dq_open(fd, O_WRONLY, dqb_path);
	// position to begin table
	dq_seek(fd, (MAGICLEN+sizeof(__u32)), dqb_path);
	// wait unlock
	dq_writelock(fd,&lock,MAGICLEN+sizeof(__u32),sizeof(__u32));
	// flush counter
	dq_write(fd, &cnt, sizeof(__u32), dqb_path);
	// wait unlock
	dq_writelock(fd,&lock,MAGICLEN+2*sizeof(__u32),cnt*DQ_ASP_SIZE);
	// flush table
	dq_write(fd, dq, cnt*DQ_ASP_SIZE, dqb_path);
	// close
	dq_close(fd, dqb_path);
	
	return 0;
}
/*
 * Calculate directory size
 * Parameters:
 *    in const char *fpath	- Path to directory
 *    in u32 type			- Type for calculate (all/uid/gid)
 *    in u32 id				- For uid/gid type - ident number
 *    in u32 rec			- Recursive lookup 
 *	  out u32 *units		- Number of memory units (bytes)
 *	  out u32 *inodes		- Number of inodes
 * Return:
 *    0			- Ok
 *    EINVAL	- Bad input values
 *    EFAULT    - Can't dir open
 */
int dir_size(in const char *fpath, in __u32 type, in __u32 id, in __u32 rec, out __u32 *units, out __u32 *inodes) {

	struct stat		stat_buf;
	DIR				*dir;
	struct dirent	*de;
	__u32			cur_units=0, rec_units=0, cur_inodes=0, rec_inodes=0, i, u;
	path			fname;
	char			*p;
	int				status, f_reiserfs;
	
	if (0 != (dir = (DIR *)opendir((const char *)fpath))) {
			
		strcpy(fname, fpath);
		// If '/' absente or not last symbol - concatenate
		p = (char *)rindex(fname, (int)'/');
		if ((0 == p) || ((0 != p) && (0 != p[1]))) strcat(fname, "/");
		// Set to end
		p = &fname[strlen(fname)];

		while (0 != (de = readdir((DIR *)dir))) {
			if ('.' == de->d_name[0]) {
				// Ignore '..'
				if (('.' == de->d_name[1]) && (0 == de->d_name[2])) continue;
			}
			*p = 0;
			strcpy(p, de->d_name);

			if (0 != lstat(fname, &stat_buf)) {
				printf("Bad lstat for %s; errno=%d\n", fname, errno);
				continue;
			}

            if (0 == (S_ISLNK(stat_buf.st_mode) || S_ISDIR(stat_buf.st_mode) || S_ISREG(stat_buf.st_mode))) continue;

            if (is_reiserfs(fname, &f_reiserfs)) {
				printf("is_reiserfs error=%d: %s is absente\n", errno, fname);
				continue;
            }
			
			if (0 == f_reiserfs) {
                // Sym links ignore
				if (S_ISLNK(stat_buf.st_mode)) continue;
				u = BLOCK_SIZE * stat_buf.st_blocks;				
				i = 1;
			}
			else {
                u = IH_SIZE + SD_SIZE + ROUND_UP(stat_buf.st_size);
                if (S_ISLNK(stat_buf.st_mode)) i = 0;
                else i = 1;
                /*
                if (S_ISLNK(stat_buf.st_mode) || S_ISDIR(stat_buf.st_mode)) {
                    //if (status = get_sd_size(fname, &sd_size)) {
                      //  printf("get_sd_size error=%d\n", status);
                      //  continue;
                    //}
					if (S_ISDIR(stat_buf.st_mode)) {
						i = 1; 
                        u = IH_SIZE + SD_SIZE + ROUND_UP(stat_buf.st_size);
					}
                    else {
                        i = 0;
                        u = IH_SIZE + SD_SIZE + ROUND_UP(stat_buf.st_size);
                    }
                }
				else {
                    //if (0 != (status = dq_asp(DQ_ASP_CMD_GETSIZE, (__u32)fname, 0, (__u32)&u))) {
                        //printf("Can't get size for %s: status=%d\n", fname, status);
                        //continue;
                    //}
                    u = IH_SIZE + SD_SIZE + ROUND_UP(stat_buf.st_size);
					i = 1;
				}
                */
			}
			
			// Reject dirs, exclude '.'
			if ((0 == S_ISDIR(stat_buf.st_mode)) || 
			(('.' == de->d_name[0]) && (0 == de->d_name[1]))) {
				switch ( type ) {
					case DQ_ASP_VEID:
						cur_units += u;
						cur_inodes += i;
						break;
					case DQ_ASP_UID:
						if (id == stat_buf.st_uid) {
                            cur_units += u;
							cur_inodes += i;
						}
						break;
					case DQ_ASP_GID:
						if (id == stat_buf.st_gid) {
                            cur_units += u;
							cur_inodes += i;
						}
						break;
					default: 
                        closedir((DIR *)dir);
                        return EINVAL;
				}
			}

			if ( rec ) {
				if (S_ISDIR(stat_buf.st_mode)) {
					// Ignore links
					if (0 == S_ISLNK(stat_buf.st_mode)) {
						// Skip '.'
						if (('.' == de->d_name[0]) && (0 == de->d_name[1])) continue;
						if (0 == (status = dir_size(fname, type, id, rec, &rec_units, &rec_inodes))) {
							cur_units += rec_units;
							cur_inodes += rec_inodes;
						}
						else {
							printf("dir_size status=%d\n", status);
                            closedir((DIR *)dir);
							return status;
						}
					}
				}
			}
		}
		*p = 0;
		closedir((DIR *)dir);
		*units = cur_units;
		*inodes = cur_inodes;
		return 0;
	}
	else {
		printf("Can't open dir %s\n", fpath);
		return EFAULT;
	}
}
/*
 * Print quotas statistics
 */
void dq_print_stat(in struct dq_stat *stat) {

	printf("all\t= %u\n", stat->all);
	printf("disable\t= %u\n", stat->disable);
	printf("veroot\t= %u\n", stat->veroot);
	printf("veid\t= %u\n", stat->type[DQ_ASP_VEID]);
	printf("uid\t= %u\n", stat->type[DQ_ASP_UID]);
	printf("gid\t= %u\n", stat->type[DQ_ASP_GID]);
}
/*
 * Flush hash table to disk
 */
void dq_sync_u(void) {

	dq_asp(DQ_ASP_CMD_SYNC, 0, 0, 1);
}
/*
 * Get type File System
 * ext2 = 0xEF53
 * reiserfs = 0x52654973
 * Parameters:
 *   in char *fpath		- Any path within mounted system
 *   out long *fstype	- Type FS 
 * Return:
 *   0		- Ok
 *   -1     - Error (see errno for detail)
 */
int	get_fstype(in char *fpath, out long *fstype) {

	struct statfs	st;
	int				ret;

	ret = statfs(fpath, &st);
	*fstype = st.f_type;
	return ret;
}
/*
 * File system is ReiserFS?
 * Parameters:
 *   in char *fpath		    - Any path within mounted system
 *   out int *flag_reiserfs - Flag - This partition is ReiserFS    
 * Return:
 *   0 - Ok
 *   1 - I/O error (more likely - this item is absente)
 */
#define REISERFS_SUPER_MAGIC	0x52654973

static int	is_reiserfs(in char *fpath, out int *flag_reiserfs) {

	long 		fstype;
	char		q[4096];

    if (conv_path(q, fpath)) return EIO;
	get_fstype(q, &fstype);
	if (REISERFS_SUPER_MAGIC == fstype) *flag_reiserfs = 1;
    else *flag_reiserfs = 0;
	return 0;
}
/*
 * Convert path for REG & LNK to parent directory path
 */
static int conv_path(char *dst, char *fpath) {

    char        *p;
	struct stat	st;

    strcpy(dst, fpath);
    if (0 != lstat(dst, &st)) {
		printf("Bad lstat for %s: errno=%d\n", fpath, errno);
		return 1;
	}
    if (S_ISDIR(st.st_mode)) return 0;

    p = (char *)rindex(dst, (int)'/');
    if (0 == p) dst[0] = 0; else p[1] = 0;
    strcat(dst, ".");
    return 0;
}

#ifdef DQLIB_DEBUG
/*
 * Print file statistics
 * Parameters:
 *    in struct stat *st		- Statistics for print
 * Return:
 *    None
 */
static void print_lstat(in struct stat *st) {

	if (S_ISDIR(st->st_mode)) printf("\tDIR\n");
	else if (S_ISLNK(st->st_mode)) printf("\tLNK\n");
	else if (S_ISREG(st->st_mode)) printf("\tFILE\n");
	else printf("\tUNKNOWN\n");

	printf("st_dev=(%d,%d) (%d)\n",	MAJOR(st->st_dev),MINOR(st->st_dev),st->st_dev);/* device */
	printf("st_ino=%d\n",		st->st_ino);      /* inode							*/
	printf("st_mode=%o\n",		st->st_mode);     /* protection						*/
	printf("st_nlink=%d\n",		st->st_nlink);    /* number of hard links			*/
	printf("st_uid=%d\n",		st->st_uid);      /* user ID of owner				*/
	printf("st_gid=%d\n",		st->st_gid);      /* group ID of owner				*/
	printf("st_rdev=%d\n",		st->st_rdev);     /* device type (if inode device)	*/
	printf("st_size=%d\n",		st->st_size);     /* total size, in bytes			*/
	printf("st_blksize=%d\n",	st->st_blksize);  /* blocksize for filesystem I/O	*/
	printf("st_blocks=%d\n",	st->st_blocks);   /* number of blocks allocated		*/

	{
		int n, b, r;

		if (S_ISDIR(st->st_mode)) {
			n = (st->st_size + 511) >> 9;
		}
		else {
			b = st->st_blksize;
			r = st->st_size % b;
			n = st->st_size / b;
			if (0 != r) n++;
			n = (n * b) >> 9;
		}

		printf("st_blocks2=%d\n", n);   /* number of blocks allocated	*/
	}
}
/*
 * Print FS statistics
 * Parameters:
 *    in char *fpath		- Any path within mounted system
 * Return:
 *    0 - Ok
 *   -1 - Error (see errno for detail)
 */
static int print_statfs(in char *fpath) {

	struct statfs	st;
	int				ret;

	statfs(fpath, &st);
/*
struct statfs {
	long    f_type;     // type of filesystem (see below)
	long    f_bsize;    // optimal transfer block size
	long    f_blocks;   // total data blocks in file system
	long    f_bfree;    // free blocks in fs
	long    f_bavail;   // free blocks avail to non-superuser
	long    f_files;    // total file nodes in file system
	long    f_ffree;    // free file nodes in fs
	fsid_t  f_fsid;     // file system id
	long    f_namelen;  // aximum length of filenames
	long    f_spare[6]; // spare for later
};
*/
	printf("f_type=0x%lx\n", st.f_type);
	printf("f_bsize=%ld\n", st.f_bsize);
	printf("f_blocks=%ld\n", st.f_blocks);
	printf("f_bfree=%ld\n", st.f_bfree);
	printf("f_bavail=%ld\n", st.f_bavail);
	printf("f_files=%ld\n", st.f_files);
	printf("f_ffree=%ld\n", st.f_ffree);
	printf("f_fsid=%ld\n", st.f_fsid);
	printf("f_namelen=%ld\n", st.f_namelen);
	return ret;
}
/*
 * Get block size for FS
 * Parameters:
 *   in char *fpath		- Any path within mounted system
 *   out int *blksize   - Block size for this partition
 * Return:
 *   0 - Ok
 *   1 - I/O error (more likely - this item is absente)
 */
static int  get_blksize(in char *fpath, out int *blksize) {

	char		q[4096];

    if (conv_path(q, fpath)) return EIO;
    return dq_asp_misc(DQ_ASP_CMD_GET_BLOCKSIZE, (unsigned)q, (unsigned)blksize, 0, 0);
}
/*
 * Get stat data for Reiser FS
 * Parameters:
 *   in char *fpath		- Any path within mounted system
 *   out int *sd_size   - Stat data size for this partition
 * Return:
 *   0 - Ok
 *   1 - I/O error (more likely - this item is absente)
 */
static int  get_sd_size(in char *fpath, out int *sd_size) {

	char		q[4096];

    if (conv_path(q, fpath)) return EIO;
    return dq_asp_misc(DQ_ASP_CMD_GET_SD_SIZE, (unsigned)q, (unsigned)sd_size, 0, 0);
}
#endif // DQLIB_DEBUG
/*
 * Ask system: dq run?
 * Return:
 *    0 - Ok
 *    1 - Not run
 */
int dq_ask(void) {
    
    int ver;

    return dq_asp(DQ_ASP_CMD_GET_VERSION, 0, 0, (unsigned)&ver);
}
/*
 * Run on string - ignore leading spaces and tabs
 */
static char *gostr(char *s) {
    
    while (*s && ((' ' == *s) || ('\t' == *s))) s++;
    return s;
}
/*
 * Check string - all symbols is digit (0..9)?
 * Return:
 *  0 - Ok
 *  1 - Bad format (any char not digit)
 */
int str_isdigit(char *s) {

    // Empty string
    if (0 == strlen(s)) return 1;

    while ( *s ) {
        if (! isdigit(*s)) return 1;
        s++;
    }
    return 0;
}
/*
 * Check string - string is dd:hh:mm:ss?
 * Return:
 *  0 - Ok
 *  1 - Bad format 
 */
static int str_istime(char *s, int n) {

    int     i=0;

    while ( *s ) {
        if (isdigit(*s) || (':' == *s)) {
            if (':' == *s) {
                i++;
                if (i > n) return 1;
            }
            s++;
        }
        else return 1;
    }
    return 0;
}
/* 
 * Convert time string to digit
 * String format:
 *  1. dd:hh:mm:ss
 *  2. xxA, where A - h/H(hour);d/D(day);w/W(week);m/M(month);y/Y(year)
 *  For instance: 7D - 7 days; 01w - 1 week; 3m - 3 months
 * Return:
 *      0 - Ok
 *      1 - Bad input params
 *      2 - Error format
 */
typedef struct dq_time {
    char        c;
    long int    t;
} dq_time;

static int dq_thr[4] = {
    1, DQ_ASP_MIN, DQ_ASP_HOUR, DQ_ASP_DAY
};

static dq_time dq_t[] = {
    { 'H', DQ_ASP_HOUR },
    { 'D', DQ_ASP_DAY },
    { 'W', DQ_ASP_WEEK },
    { 'M', DQ_ASP_MONTH },
    { 'Y', DQ_ASP_YEAR },
    { 0, 0 }
};

#define ASP_STR_LEN     256

int dq_str2time(char *s, long int *n) {
    
    int         len, i, tmp;
    long int    val=0;
    char        c, buf[ASP_STR_LEN], *p, *q;

    // Check input params 
    if ((0 == s) || (0 == n)) return DQ_ASP_BADIN;
    if (0 == (len = strlen(s))) return DQ_ASP_BADIN;
    if (len >= ASP_STR_LEN) return DQ_ASP_BADIN;

    c = s[len-1];   // last char (Q: trail spaces?)
    strncpy(buf, s, ASP_STR_LEN);

    if (isdigit(c)) {
        // Format 1
        // Ignore leading spaces & tabs
        p = gostr(buf);

        // Only seconds?
        if (0 == str_isdigit(p)) {
            if (0 == strlen(p)) return DQ_ASP_BADFORMAT;
            sscanf(p, "%ld", n);
            return 0;
        }

        if (str_istime(p, 3)) return DQ_ASP_BADFORMAT;     // Bad format
        for (i = 0; i < 4; i++) {
            q = (char *)rindex(p, (int)':');
            if ( q ) {
                *q++ = 0;
                // Check - may be empty field
                if (0 == strlen(q)) return DQ_ASP_BADFORMAT;
                sscanf(q, "%u", &tmp);
            }
            else {
                // Check - may be empty field
                if (0 == strlen(p)) return DQ_ASP_BADFORMAT;   
                sscanf(p, "%u", &tmp);
            }
            // Convert to seconds
            tmp *= dq_thr[i];
            if (i < 3) {
                // Check overflow
                if (tmp >= dq_thr[i+1]) return DQ_ASP_BADFORMAT;
            }
            val += tmp;
            if (0 == q) { *n = val; return 0; }
        }
        *n = val;
    }
    else {
        // Format 2
        // Convert to uppercase
        c = toupper(c);
        // Reject last char
        buf[len-1] = 0; 
        // Look up
        for (i = 0; dq_t[i].c; i++) {
            if (c == dq_t[i].c) {
                // Ignore leading spaces & tabs
                p = gostr(buf);           
                if (str_isdigit(p)) return DQ_ASP_BADFORMAT;   // Bad format
                sscanf(p, "%ld", &val);
                *n = dq_t[i].t * val; 
                return 0;
            }
        };
        return DQ_ASP_BADFORMAT;   // Bad format
    }
    return 0;
}
/*
 * Convert digital time (seconds) to string
 * String format:
 *  1. dd:hh:mm:ss
 *  2. xxA, where A - h/H(hour);d/D(day);w/W(week);m/M(month);y/Y(year)
 *  For instance: 7D - 7 days; 01w - 1 week; 3m - 3 months
 * Return:
 *      0 - Ok
 *      1 - Bad input params
 */
int dq_time2str(long int n, char *s, int format) {

    int i;
    int sec=0,min=0,hour=0; 
    long int t,day=0;

    // Check input params 
    if (0 == s) return DQ_ASP_BADIN;

    if (0L == n) { s[0] = '0'; s[1] = 0; return 0; }

    if ( format ) {
        for (i = 4; i >= 0; i--) {
            if (0 == (n % dq_t[i].t)) {
                sprintf(s, "%u%c", (unsigned)n / (unsigned)dq_t[i].t, dq_t[i].c);
                return 0;
            }
        }
        // Set dd:hh:mm:ss
        t = n;
        sec = t % 60;

        t /= 60;
        min = t % 60;

        t /= 60;
        hour = t % 24;

        day = t / 24;

        if ( day ) sprintf(s, "%lu:%02u:%02u:%02u", day, hour, min, sec);
        else if ( hour ) sprintf(s, "%02u:%02u:%02u", hour, min, sec);
        else if ( min ) sprintf(s, "%02u:%02u", min, sec);
        else sprintf(s, "%02u", sec);
    }
    else sprintf(s, "%lu", n);        

    return 0;
}
/*
 * Convert string to units
 * String format:
 * 1. xxK (xxk) - size in Kilobytes
 * 2. xxM (xxm) - size in Megabytes
 * 3. bytes
 * Return:
 *  0 - Ok
 *  1 - Bad input params
 *  2 - Bad format
 */
int dq_str2unit(char *s, long int *n) {

    int         len;
    char        c, buf[ASP_STR_LEN], *p;

    // Check input params 
    if ((0 == s) || (0 == n)) return DQ_ASP_BADIN;
    if (0 == (len = strlen(s))) return DQ_ASP_BADIN;
    if (len >= ASP_STR_LEN) return DQ_ASP_BADIN;

    c = s[len-1];   // last char (Q: trail spaces?)
    strncpy(buf, s, ASP_STR_LEN);
    // Ignore leading spaces & tabs
    p = gostr(buf);
    if (0 == strlen(p)) return DQ_ASP_BADFORMAT;

    if (isdigit(c)) {
        if (0 == str_isdigit(p)) {
            sscanf(s, "%ld", n);
            return 0;
        }
        return DQ_ASP_BADFORMAT;
    }
    // Convert to uppercase
    c = toupper(c);
    // Reject last char
    buf[len-1] = 0; 
    if (0 == strlen(p)) return DQ_ASP_BADFORMAT;
    if (0 != str_isdigit(p)) return DQ_ASP_BADFORMAT;

    sscanf(s, "%ld", n);

    switch ( c ) {
        case 'K': *n *= 1024; break;
        case 'M': *n *= 1024*1024; break;
        default: return DQ_ASP_BADFORMAT;
    }
    return 0;
}
/*
 * Convert digital units to string
 * String format:
 *  1. xxK - whole Kb
 *  2. xxM - whole Mb
 *  3. xxx - bytes only
 *  For instance: 30M - 30 Mbytes; 100K - 100 Kbytes
 * Return:
 *      0 - Ok
 *      1 - Bad input params
 */
int dq_unit2str(long int n, char *s, int format) {

    // Check input params 
    if (0 == s) return DQ_ASP_BADIN;

    if (0L == n) { s[0] = '0'; s[1] = 0; return 0; }

    if ( format ) {
        if (0 == (n % MB)) {
            n /= MB;
            sprintf(s, "%ldM", n);
        }
        else if (0 == (n % KB)) {
            n /= KB;
            sprintf(s, "%ldK", n);
        }
        else sprintf(s, "%ld", n);
    }
    else sprintf(s, "%ld", n);

    return 0;
}
/*
 * Put quota structure to kernel
 * Parameters:
 *    struct dq_blk *dq       - Ptr to quota structure
 *    const char *dqb_path    - Path to data base file
 *    int offset              - Offset to data base
 * Return:
 *    0      - Ok
 *    EINVAL - Bad input params 
 */
int dq_put(struct dq_blk *dq, const char *dqb_path, int offset) {

    int status;

    // Check input params
    if ((0 == dq) || (0 == dqb_path)) return EINVAL;
    if (offset < 0) return EINVAL;
    // Reset dirty flag
    dq->flags &= ~DQ_ASP_DIRTY;
    // Put to kernel (syscall dq_asp)
    if (0 != (status = dq_asp(DQ_ASP_CMD_PUT, (__u32)dq, (__u32)dqb_path, (__u32)offset))) {
	    printf("\nBad status=%d\n", status);
	    printf("(veid,uid,guid,flags)=(%u,%u,%u,%x)\n", dq->veid, dq->uid, dq->gid, dq->flags);
    }
    return status;
}
/*
 * Put quota structure to kernel
 * Parameters:
 *    struct dq_blk *dq       - Ptr to quota structure
 *    const char *dqb_path    - Path to data base file
 *    int index               - Index to data base
 * Return:
 *    0      - Ok
 *    EINVAL - Bad input params 
 */
int dq_put_index(struct dq_blk *dq, const char *dqb_path, int index) {

    return dq_put(dq, dqb_path, DQB_HDRSIZE+index*sizeof(struct dq_blk));
}
