/*
*
* Author:              Vadim V. Sinitsyn
* E-mail:              sinvv@asplinux.ru
*
* $Header: /home/cvs/ASPcomplete/utils/initquota.c,v 1.10 2000/10/18 08:14:54 sinvv Exp $
*
* Remarks:
*
* Last correct: 18/Oct/00
*
*/
#include <stdio.h>
#include <string.h>
#include <pwd.h>
#include <sys/types.h>
#include "dqlib.h"

static void	usage(in const char *task);

static path		dqb_path = DEFAULT_DQB,
				init_path = DQ_ASP_VE_ROOT,
				id_path;

static struct passwd    *pwd;

static int		n_ve;
static int		a_flag=0, 
				i_flag=0, 
				f_flag=0, 
				p_flag=0,
				d_flag=0,
				u_flag=0, g_flag=0, v_flag=0;

int main(int argc, char **argv) {

	int				ret, status;
	__u32			i, cnt, ver, id, units=0, inodes=0;
	__u32			type;
	struct dq_blk	*dq, *q;

	// Check policy
	check_root;
	// Set ve number
	if (0 == is_ve0()) n_ve = VEID_0; else n_ve = my_veid;
    // Kernel compile with ASP Disk Quota?
    DQ_ASP_IS;

	while ((ret = getopt(argc, argv, "haf:v:u:g:pi:D")) != EOF) {
		switch ( ret ) {
			case 'h':
			case '?':
					usage(argv[0]);
					return 1;
			case 'a':
					a_flag = 1;
					break;
			case 'p':
					p_flag = 1;
					break;
			case 'f':
					f_flag = 1;
					strncpy(dqb_path, optarg, PATH_SIZE-1);
					dqb_path[PATH_SIZE-1] = 0;
					break;
			case 'i':
					i_flag = 1;
					strncpy(init_path, optarg, PATH_SIZE-1);
					init_path[PATH_SIZE-1] = 0;
					break;
			case 'u':
					u_flag = 1;
                    // Check string: only decimal digits?
                    if (str_isdigit(optarg)) {
                        // Try found user with mnemonic name
                        pwd = getpwnam(optarg);
                        if (0 == pwd) {
                            printf("User %s not exist\n", optarg);
                            exit(1);
                        }
                        // Success
                        id = pwd->pw_uid;
                    }
                    else sscanf(optarg,"%u", &id);
					break;
			case 'g':
					g_flag = 1;
                    sscanf(optarg,"%u", &id);
					break;
			case 'v':
					v_flag = 1;
                    sscanf(optarg,"%u", &id);
					break;
            case 'D':
                    d_flag = 1;
                    break;
		}
	}
    // Debug print
    if ( d_flag ) {
        printf("a_flag=%d;p_flag=%d;u_flag=%d;g_flag=%d;v_flag=%d;id=%d;f_flag=%d;dqb_path=%s;i_flag=%d;init_path=%s\n",
            a_flag,p_flag,u_flag,g_flag,v_flag,id,f_flag,dqb_path,i_flag,init_path);
    }

	// Check flags consistency
	if (1 < (a_flag + p_flag)) usage(argv[0]);

    // Check flags consistency
	if (1 != (u_flag + g_flag + v_flag)) usage(argv[0]);

    // Set init path
    if (0 == i_flag) {
        // If init path not set 
        if ( u_flag ) {
            pwd = getpwuid(id);
            if (0 == pwd) {
                printf("User with uid=%d not exist\n", id);
                exit(1);
            }
            else {
                strcpy(id_path, pwd->pw_dir);
                if ( d_flag ) printf("pw_dir=%s\n",id_path);
            }
        }
        else {
            if (0 == v_flag) usage(argv[0]);
            sprintf(id_path, "%s%d", init_path, id);
        }
    }
    else strcpy(id_path, init_path);

    // Set dqb_path
    if (0 == f_flag) {
        // Path to data base not defined explicitly 
        char *p;
        if (0 != (p = getenv(PATH_DQB_ENV))) {
            strcpy(dqb_path, p);
            if ( d_flag ) printf("dqb_path=%s\n", dqb_path);
        }
    }

    // Set type
    if ( v_flag ) type = DQ_ASP_VEID;
    else if ( u_flag ) type = DQ_ASP_UID;
    else type = DQ_ASP_GID;

    // Flush current state to disk
    dq_sync_u();

    if (0 == p_flag) {
		// Read dqb 
		if (0 != (status = read_dqb(dqb_path, &dq, &cnt, &ver))) {
			printf("Can't read dqb %s: status error=%d\n", dqb_path, status);	
			return status;
		}
	}
    else {
		if (0 != (status = dir_size(id_path, type, id, 1, &units, &inodes))) {
			printf("Bad status dir_size: %d\n", status);	
		}
		else printf("units=%u;inodes=%u\n", units, inodes);
        return 0;
	}

    for (i = 0, q = dq; i < cnt; i++, q++) {
		// Ignore deleted records
		if (DQ_ASP_IS_REMOVE(q->flags)) continue;
		// Check type record
		if (type != (q->flags & DQ_ASP_VEGUID)) continue;
		// Types now at the same
		switch ( type ) {
			case DQ_ASP_VEID:
				if (id != q->veid) break;
				if (0 != (status = dir_size(id_path, type, id, 1, &units, &inodes))) {
					printf("Bad status dir_size: %d\n", status);	
				}
				else {
					if ( a_flag ) {
						q->inode_current	+= inodes;
						q->unit_current		+= units;
					}
					else {
						q->inode_current	= inodes;
						q->unit_current		= units;
					}
                    dq_put_index(q, dqb_path, i);
                    
				}
				break;
			case DQ_ASP_UID:
				if ((id != q->uid) || (n_ve != q->veid)) break;
				if (0 != (status = dir_size(id_path, type, id, 1, &units, &inodes))) {
					printf("Bad status dir_size: %d\n", status);	
				}
				else {
					if ( a_flag ) {
						q->inode_current	+= inodes;
						q->unit_current		+= units;
					}
					else {
						q->inode_current	= inodes;
						q->unit_current		= units;
					}
                    dq_put_index(q, dqb_path, i);
				}
				break;
			case DQ_ASP_GID:
				if ((id != q->gid) || (n_ve != q->veid)) break;
				if (0 != (status = dir_size(id_path, type, id, 1, &units, &inodes))) {
					printf("Bad status dir_size: %d\n", status);	
				}
				else {
					if ( a_flag ) {
						q->inode_current	+= inodes;
						q->unit_current		+= units;
					}
					else {
						q->inode_current	= inodes;
						q->unit_current		= units;
					}
                    dq_put_index(q, dqb_path, i);
				}
				break;
		}
	}

	// If hit to table - flush table on the disk
	status = flush_tbl2dqb(dqb_path, dq, cnt);
	// Free table
	dq_free(dq);
    // Debug print
    if ( d_flag ) printf("units=%u; inodes=%u\n", units, inodes);
	return 0;	
}
/*
 * Print usage 
 * Parameters:
 *    in const char *task  - Task name
 * Return:
 *    None
 */
static void usage(in const char *task) {

	if (VEID_0 == n_ve) printf("Usage: %s [-hapfivug]\n", task);
	else printf("Usage: %s [-hapfiug]\n", task);

	printf("\t-h\t\thelp for more information\n");
	printf("\t-a\t\tadd current value to dqb\n");
	printf("\t-p\t\tprint current quota value\n");
	printf("\t-f file\t\tpath to dqb file (default: /var/quota/dqb)\n");
	printf("\t-i path\t\tset path for init quota (default for veid: /ve_private/n, where n (decimal) - veid number)\n");

	printf("\t-u id\t\tuid\n");
	printf("\t-g id\t\tgid\n");
	if (VEID_0 == n_ve) printf("\t-v id\t\tveid\n");

	exit(0);
}

