/*
*
* Author:              Vadim V. Sinitsyn
* E-mail:              sinvv@asplinux.ru
*
* $Header: /home/cvs/ASPcomplete/utils/own.c,v 1.2 2000/10/17 17:08:45 sinvv Exp $
*
* Remarks:
*
* Last correct: 17/Oct/00
*
*/
#include <stdio.h>
#include "dqlib.h"

static void	usage(in const char *task);
static void print_ve_quota(in __u32 ve, in __u32 unit_out_format);

static	int d_flag=0,
			Kb_flag=0, Mb_flag=0, 
            unit_out_format=0;

static int	n_ve;

int main(int argc, char **argv) {

	int				ret;

	// Check policy
	check_root;
	// Set ve number
    if (0 == is_ve0()) {
        n_ve = VEID_0; 
        printf("VE=0 not have ASP Disk Quota\n");
        return 1;
    }
    else n_ve = my_veid;
    // Kernel compile with ASP Disk Quota?
    DQ_ASP_IS;

	while ((ret = getopt(argc, argv, "hkKmMD")) != EOF) {
		switch ( ret ) {
			case 'h':
			case '?':
					usage(argv[0]);
					return 1;
			case 'k':
			case 'K':
					Kb_flag = 1;
					unit_out_format = DQ_ASP_FORMAT_UNIT_KB;
					break;
			case 'm':
			case 'M':
					Mb_flag = 1;
					unit_out_format = DQ_ASP_FORMAT_UNIT_MB;
					break;
            case 'D':
                    d_flag = 1;
                    break;
		}
	}

    // Debug print
    if ( d_flag ) {
        printf("Kb_flag=%d;Mb_flag=%d\n",Kb_flag,Mb_flag);
    }

    // Check flag consistency
	if (1 < (Kb_flag + Mb_flag)) usage(argv[0]);

    // Flush to disk
	dq_sync_u();

    // Print data
	print_ve_quota(n_ve, unit_out_format);

	return 0;
}
/*
 * Usage message
 * Parameters:
 *    in const char *task  - Task name
 * Return:
 *    None
 */
static void usage(in const char *task) {

	if (VEID_0 == n_ve) printf("Usage: %s [-km]\n", task);
	else printf("Usage: %s [-km]\n", task);

	printf("\t-k (or -K)\tprint quota units in Kbytes (default: in bytes)\n");
	printf("\t-m (or -M)\tprint quota units in Mbytes (default: in bytes)\n");
	exit(0);
}
/*
 * Print disk quota for current ve
 * Parameters:
 *    in __u32 ve				- Number for filter 
 *    in __u32 unit_out_format	- Output format for unit (Bytes/Kb/Mb)
 * Return:
 */
static void print_ve_quota(in __u32 ve, in __u32 unit_out_format) {

	int				status;
    struct dq_blk	*q, dq = { 0 };


    dq.veid = ve;
    dq.flags = DQ_ASP_VEID;
	if (0 != (status = dq_asp(DQ_ASP_CMD_GET, (__u32)&dq, 0, 0))) {
	    //printf("Bad DQ_ASP_CMD_GET: status=%d\n", status);
		return;
	}
    q = &dq;
	// If mark to delete - ignore
	if (DQ_ASP_IS_REMOVE(q->flags)) return;
    print_dqblk_header();
    print_dqblk(q, DQ_ASP_FORMAT_LINE, unit_out_format);
}
