/*
*
* Author:              Vadim V. Sinitsyn
* E-mail:              sinvv@asplinux.ru
*
* $Header: /home/cvs/ASPcomplete/utils/quota.c,v 1.20 2000/10/18 08:14:39 sinvv Exp $
*
* Remarks:
*
* Last correct: 18/Oct/00
*
*/
#include <stdio.h>
#include <string.h>
#include "dqlib.h"

//#define DQ_TEST

static void		usage(in const char *task);
static int		set_flag(in const char *dqb_path, in __u32 id, in __u32 type, in __u32 flags, in __u32 action);
/*
* groot:
*	1. uid-type - user from veid 0 (former)
*	2. gid-type - group from veid 0 (former)
*	3. veid-type - (!=0-!!!) - for appropriate VE
* back door:
*	4: if designate uid or/and gid - groot possible create private (VE) restriction,
*	namely (veid,uid,gid) (veid!=0)
* veroot
*	1. uid-type - user from native veid (veid !=0)
*	2. gid-type - group from native veid (veid !=0)
*	3. veid-type - must be reject
*/
static	path	dqb_path = DEFAULT_DQB;

static	int f_flag=0, 
			v_flag=0, 
            g_flag=0, 
            u_flag=0,
            d_flag=0,
			disable_flag=0, 
            enable_flag=0, 
            remove_flag=0,
			s_flag=0;

#ifdef DQ_TEST
static int x_flag=0,y_flag=0,z_flag=0;
static path test_str;
#endif

static struct dq_stat	qstat;

static int	n_ve, id=0;

int main(int argc, char **argv) {

	int		        ret;
	__u32	        type=0;
	int		        status;
    struct dq_blk   dq = { 0 };

	// Check policy
	check_root;
	// Set ve number
	if (0 == is_ve0()) n_ve = VEID_0; else n_ve = my_veid;
    // Kernel compile with ASP Disk Quota?
    DQ_ASP_IS;

#ifdef DQ_TEST
    while ((ret = getopt(argc, argv, "hf:edrv:u:g:sDx:y:z:")) != EOF) {
#else
	while ((ret = getopt(argc, argv, "hf:edrv:u:g:sD")) != EOF) {
#endif
		switch ( ret ) {
			case 'h':
			case '?':
					usage(argv[0]);
					return 1;
			case 's': 
					s_flag = 1; 
					break;
			case 'd': 
					disable_flag = 1; 
					break;
			case 'e': 
					enable_flag = 1; 
					break;
			case 'r': 
					remove_flag = 1; 
					break;
			case 'f': 
					f_flag = 1; 
					strncpy(dqb_path, optarg, PATH_SIZE-1);
					dqb_path[PATH_SIZE-1] = 0;
					break;
			case 'v': 
					// VEroot not allowed operation with VE!
					if (VEID_0 != n_ve) usage(argv[0]);
					v_flag = 1;
					dq.flags = type = DQ_ASP_VEID;
                    sscanf(optarg, "%u", &id);
                    dq.veid = id;
					break;
			case 'u': 
					u_flag = 1;
					dq.flags = type = DQ_ASP_UID;
                    sscanf(optarg, "%u", &id);
                    dq.veid = n_ve;
                    dq.uid = id;
					break;
			case 'g': 
					g_flag = 1;
					dq.flags = type = DQ_ASP_GID;
                    sscanf(optarg, "%u", &id);
                    dq.veid = n_ve;
                    dq.gid = id;
					break;
            case 'D':
                    d_flag = 1;
                    break;
#ifdef DQ_TEST
            case 'x':
                    x_flag = 1;
                    strcpy(test_str, optarg);
                    break;
            case 'y':
                    y_flag = 1;
                    strcpy(test_str, optarg);
                    break;
            case 'z':
                    z_flag = 1;
                    strcpy(test_str, optarg);
                    break;
#endif
		}
	}
#ifdef DQ_TEST
{
    __u32   kdev;
    path    mnt, dev_name;
    int     i_num, n, ret;

    switch (x_flag+y_flag+z_flag) {
        case 0: break;
        case 1: 
            if ( x_flag ) {
                if (0 == dq_asp_misc(DQ_ASP_CMD_GET_KDEV, (__u32)test_str, (__u32)&kdev, (__u32)mnt, (__u32)dev_name)) {
                    printf("path=%s;dev=(%d,%d);mnt=%s;device=%s\n", test_str, (kdev >> 8) & 0xff, kdev & 0xff, mnt, dev_name);
                }
            }
            else if ( y_flag ) {
                if (0 == dq_asp_misc(DQ_ASP_CMD_GET_INODE, (__u32)test_str, (__u32)&i_num, 0, 0)) {
                    printf("path=%s;inode=%d\n", test_str, i_num);
                }
            }
            else if ( z_flag ) {
                if (0 != (ret = dq_asp(DQ_ASP_CMD_GETSIZE, (__u32)test_str, (__u32)&n, 0))) {
                    printf("dq_asp syscall error: ret=%d\n", ret);
                }
                else printf("file %s has size %d bytes\n", test_str, n);
            }
            return 1;
        default: usage(argv[0]);
    }
}
#endif
    // Debug print
    if ( d_flag ) {
        printf("u_flag=%d;g_flag=%d;v_flag=%d;id=%d;f_flag=%d;dqb_path=%s;disable_flag=%d;enable_flag=%d;remove_flag=%d\n",
            u_flag,g_flag,v_flag,id,f_flag,dqb_path,disable_flag,enable_flag,remove_flag);
    }

	if ( s_flag ) {
		if (enable_flag | disable_flag | remove_flag | v_flag | u_flag | g_flag) usage(argv[0]);
		if (0 == (status = dq_asp(DQ_ASP_CMD_GETSTAT, (__u32)&qstat, 0, sizeof(struct dq_stat)))) {
			dq_print_stat(&qstat);
		}
		else printf("Bad status get statistics: %d (%d)\n", status, DQ_ASP_CMD_GETSTAT);
		return status;
	}
	// Check flags
	if (1 != (enable_flag + disable_flag + remove_flag)) usage(argv[0]);
	if (1 != (v_flag + u_flag + g_flag)) usage(argv[0]);

    // Set dqb_path
    if (0 == f_flag) {
        // Path to data base not defined explicitly 
        char *p;
        if (0 != (p = getenv(PATH_DQB_ENV))) {
            strcpy(dqb_path, p);
            if ( d_flag ) printf("dqb_path=%s\n", dqb_path);
        }
    }

    if ( enable_flag ) {
        dq_asp(DQ_ASP_CMD_ENABLE, (__u32)&dq, 0, 0);
        status = set_flag(dqb_path, id, type, DQ_ASP_DISABLE, DQ_ASP_OFF);
    }
    else if ( disable_flag ) {
        dq_asp(DQ_ASP_CMD_DISABLE, (__u32)&dq, 0, 0);
        status = set_flag(dqb_path, id, type, DQ_ASP_DISABLE, DQ_ASP_ON);
    }
    else {
        dq_asp(DQ_ASP_CMD_REMOVE, (__u32)&dq, 0, 0);
        status = set_flag(dqb_path, id, type, DQ_ASP_REMOVE, DQ_ASP_ON);
    }

	return status;
}
/*
 * Usage message
 * Parameters:
 *    in const char *task  - Task name
 * Return:
 *    None
 */
static void usage(in const char *task) {

	if (VEID_0 == n_ve) printf("Usage: %s [-hfedrvug]\n", task);
	else printf("Usage: %s [-hfedrug]\n", task);

	printf("\t-s\t\tprint current quota statistics\n");
	printf("\t-f file\t\tpath to dqb file (default: /var/quota/dqb)\n");
	printf("\t-e\t\tenable appropriate quota\n");
	printf("\t-d\t\tdisable appropriate quota\n");
	printf("\t-r\t\tremove appropriate quota\n");
	
	printf("\t-u n\t\tquota for users with uid number n\n");
	printf("\t-g n\t\tquota for groups with gid number n\n");
	if (VEID_0 == n_ve) printf("\t-v n\t\tquota for VE with veid number n\n");
	
	exit(0);
}
/*
 * Set/reset dqb flags
 * Parameters:
 *    in const char *dqb_path	- Path to dq data base file
 *    in __u32 id				- id
 *    in __u32 type				- type id (uid,gid,veid) 
 *    in __u32 flags			- Flags for set/reset
 *    in __u32 action			- Action 0-reset/1-set
 * Return:
 *    0					- Ok
 *    EINVAL			- NULL ptr to dqb
 *    ESRCH				- File not found (see errno for more detail)
 *    EACCES			- Bad rights
 *    EIO				- Bad open/read/seek or close
 */
static int set_flag(
	in const char *dqb_path, 
	in __u32 id,
	in __u32 type,
	in __u32 flags,
	in __u32 action) {

	int				status;
	__u32			cnt, ver, i;
	struct dq_blk	*dqb, *q;

	// Check input params
	if (NULL == dqb_path) return EINVAL;

    // Flush current state to disk
    dq_sync_u();

	if (0 != (status = read_dqb(dqb_path, &dqb, &cnt, &ver))) return status;

	for (i = 0, q = dqb; i < cnt; i++, q++) {
		// If mark to delete - ignore
		if (DQ_ASP_IS_REMOVE(q->flags)) continue;
		// Check type
		if (DQ_ASP_MAXTYPE <= (q->flags & DQ_ASP_VEGUID)) continue;

		switch (q->flags & DQ_ASP_VEGUID) {
			case DQ_ASP_UID:
				if (q->uid == id) {
					if (DQ_ASP_ON == action) q->flags |= flags;
					else q->flags &= ~flags;
				}
				break;
			case DQ_ASP_GID:
				if (q->gid == id) {
					if (DQ_ASP_ON == action) q->flags |= flags;
					else q->flags &= ~flags;
				}
				break;
			case DQ_ASP_VEID:
				// This operation allowed only for groot
				if (VEID_0 != n_ve) return EACCES;
				if (q->veid == id) {
					if (DQ_ASP_ON == action) q->flags |= flags;
					else q->flags &= ~flags;
				}
				break;
			default: return EFAULT;
		}
	}

    status = flush_tbl2dqb(dqb_path, dqb, cnt);

	// Free table
	dq_free(dqb);
	return status;
}
