/*
*
* Author:              Vadim V. Sinitsyn
* E-mail:              sinvv@asplinux.ru
*
* $Header: /home/cvs/ASPcomplete/utils/quota_add.c,v 1.6 2000/10/18 08:14:45 sinvv Exp $
*
* Remarks:
*
* Last correct: 18/Oct/00
*
*/
#include <stdio.h>
#include <string.h>
#include "dqlib.h"

typedef enum asp_state {
    ASP_UCURRENT=1,
    ASP_USOFT,
    ASP_UHARD,
    ASP_UTIME,
    ASP_UEXPIRATION,
    ASP_ICURRENT,
    ASP_ISOFT,
    ASP_IHARD,
    ASP_ITIME,
    ASP_IEXPIRATION
} asp_state;

#define ASP_TYPE_TIME       0x80

static void		usage(in const char *task);
static int		dq_append(in const char *dqb_path, in struct dq_blk *dq);
static int      add_dqb(in const char *dqb_path, in struct dq_blk *p);
/*
* groot:
*	1. uid-type - user from veid 0 (former)
*	2. gid-type - group from veid 0 (former)
*	3. veid-type - (!=0-!!!) - for appropriate VE
* back door:
*	4: if designate uid or/and gid - groot possible create private (VE) restriction,
*	namely (veid,uid,gid) (veid!=0)
* veroot
*	1. uid-type - user from native veid (veid !=0)
*	2. gid-type - group from native veid (veid !=0)
*	3. veid-type - must be reject
*/

static	path	dqb_path = DEFAULT_DQB;

static	int u_flag=0,g_flag=0,v_flag=0, 
			s_flag=0,
            d_flag=0,
            f_flag=0,
			type=0,id=0;

static int	n_ve;

int main(int argc, char **argv) {

	int				ret;
    struct dq_blk   dq = { 0 };
    int             opt_index=0,flag=0;
    long int        t, val;

    // uc us uh uct uet ic is ih ict iet
    struct option opt[] = {
        { "ucurrent",   1, &flag, ASP_UCURRENT },        // unit current
        { "usoft",      1, &flag, ASP_USOFT },           // unit soft limit
        { "uhard",      1, &flag, ASP_UHARD },           // unit hard limit
        { "utime",      1, &flag, ASP_UTIME | ASP_TYPE_TIME },      // unit current time (after soft threshold)
        { "uexpiration",1, &flag, ASP_UEXPIRATION | ASP_TYPE_TIME}, // unit expiration time (after soft threshold)
        { "icurrent",   1, &flag, ASP_ICURRENT },        // inode current
        { "isoft",      1, &flag, ASP_ISOFT },           // inode soft limit
        { "ihard",      1, &flag, ASP_IHARD },           // inode hard limit
        { "itime",      1, &flag, ASP_ITIME | ASP_TYPE_TIME},       // inode current time (after soft threshold)
        { "iexpiration",1, &flag, ASP_IEXPIRATION | ASP_TYPE_TIME}, // inode expiration time (after soft threshold)
        { 0, 0, 0, 0 }
    };

	// Check policy
	check_root;
	// Set ve number
	if (0 == is_ve0()) n_ve = VEID_0; else n_ve = my_veid;
    // Kernel compile with ASP Disk Quota?
    DQ_ASP_IS;
	
	while ((ret = getopt_long(argc, argv, "hsf:v:u:g:D", opt, &opt_index)) != EOF) {
		switch ( ret ) {
            case 0:
                // Check and convert input params 
                if (flag & ASP_TYPE_TIME) {
                    if (dq_str2time(optarg, &t)) { printf("bad format: %s)\n", optarg); return 0; }
                    flag &= ~ASP_TYPE_TIME;
                }
                else {
                    if ((ASP_UCURRENT == flag) || (ASP_USOFT == flag) || (ASP_UHARD == flag)) {
                        if (dq_str2unit(optarg, &val)){ printf("bad format: %s\n", optarg); return 0; }
                    }
                    else {
                        if (str_isdigit(optarg)) { printf("bad format: %s\n", optarg); return 0; }
                        sscanf(optarg, "%ld", &val);
                    }
                }

                switch ( flag ) {
                    case ASP_UCURRENT:
                        dq.unit_current = val;
                        break;
                    case ASP_USOFT:
                        dq.unit_soft = val;
                        break;
                    case ASP_UHARD:
                        dq.unit_hard = val;
                        break;
                    case ASP_UTIME:
                        dq.cur_unit_time = (time_t)t;
                        break;
                    case ASP_UEXPIRATION:
                        dq.unit_expiration = (time_t)t;
                        break;
                    case ASP_ICURRENT:
                        dq.inode_current = val;
                        break;
                    case ASP_ISOFT:
                        dq.inode_soft = val;
                        break;
                    case ASP_IHARD:
                        dq.inode_hard = val;
                        break;
                    case ASP_ITIME:
                        dq.cur_inode_time = (time_t)t;
                        break;
                    case ASP_IEXPIRATION:
                        dq.inode_expiration = (time_t)t;
                        break;
                    default: 
                        break;
                };
                break;
			case 'h':
			case '?':
				usage(argv[0]);
				return 1;
			case 's': 
					s_flag = 1; 
					break;
			case 'f': 
					strncpy(dqb_path, optarg, PATH_SIZE-1);
					dqb_path[PATH_SIZE-1] = 0;
					break;
			case 'v': 
					// VEroot not allowed operation with VE!
					if (VEID_0 != n_ve) usage(argv[0]);
					v_flag = 1; 
					dq.flags = type = DQ_ASP_VEID;
                    sscanf(optarg,"%u", &id);
                    dq.veid = dq.luid = id; 
					break;
			case 'u': 
					u_flag = 1; 
					dq.flags = type = DQ_ASP_UID;
                    sscanf(optarg,"%u", &id);
                    dq.uid = id; 
                    dq.veid = dq.luid = n_ve;
					break;
			case 'g': 
					g_flag = 1;
					dq.flags = type = DQ_ASP_GID;
                    sscanf(optarg,"%u", &id);
                    dq.gid = id; 
                    dq.veid = dq.luid = n_ve;
					break;
            case 'D':
                    d_flag = 1;
                    break;
		}
	}

    // Debug print
    if ( d_flag ) {
        printf("u_flag=%d;g_flag=%d;v_flag=%d;id=%d;f_flag=%d;dqb_path=%s\n",
            u_flag,g_flag,v_flag,id,f_flag,dqb_path);
    }

	// Check flags consistency
	if (1 != (s_flag + v_flag + u_flag + g_flag)) usage(argv[0]);

    // Set dqb_path
    if (0 == f_flag) {
        // Path to data base not defined explicitly 
        char *p;
        if (0 != (p = getenv(PATH_DQB_ENV))) {
            strcpy(dqb_path, p);
            if ( d_flag ) printf("dqb_path=%s\n", dqb_path);
        }
    }

    // Set empty data base
    if ( s_flag ) {
		create_empty_dqb(dqb_path);
		return 0;
	}

    // Check settings 
    if (dq.unit_soft > dq.unit_hard) { printf("bad params: soft limit>hard limit\n"); return 1; }
    if (dq.inode_soft > dq.inode_hard) { printf("bad params: soft limit>hard limit\n"); return 1; }

    // Set default expiration time
    if (0 == dq.unit_expiration) dq.unit_expiration = DEFAULT_EXPIRATION;
    if (0 == dq.inode_expiration) dq.inode_expiration = DEFAULT_EXPIRATION;

    // Debug print
    if ( d_flag ) {
        print_dqblk(&dq, DQ_ASP_FORMAT_COLUMN, 0);
    }

	add_dqb(dqb_path, &dq);
	return 0;
}
/*
 * Usage message
 * Return:
 *    None
 */
static void usage(const char *task) {

	if (VEID_0 == n_ve) printf("Usage: %s [-hsfvug --uc --us --uh --ut --ue --ic --is --ih --it --ie]\n", task);
	else printf("Usage: %s [-hsfug --uc --us --uh --ut --ue --ic --is --ih --it --ie]\n", task);

	printf("\t-s\t\t\tcreate new dqb\n");
	printf("\t-f\t\t\tpath to dqb file (default: /var/quota/dqb)\n");

    printf("\t-u id\t\t\tadd to dqb quotas for users\n");
	printf("\t-g id\t\t\tadd to dqb quotas for groups\n");
	if (VEID_0 == n_ve) printf("\t-v id\t\t\tadd to dqb quotas for VE\n");

	printf("\t--uc[urrent] n\t\tunit current fot quota\n");
    printf("\t--us[oft] n\t\tunit soft limit fot quota\n");
    printf("\t--uh[ard] n\t\tunit hard limit fot quota\n");
    printf("\t--ut[ime] n\t\tunit current time fot quota\n");
    printf("\t--ue[xpiration] n\tunit expiration time fot quota\n");

    printf("\t--ic[urrent] n\t\tinode current fot quota\n");
    printf("\t--is[oft] n\t\tinode soft limit fot quota\n");
    printf("\t--ih[ard] n\t\tinode hard limit fot quota\n");
    printf("\t--it[ime] n\t\tinode current time fot quota\n");
    printf("\t--ie[xpiration] n\tinode expiration time fot quota\n");

	exit(0);
}
/*
 * Append quota to end of dqb file
 * Parameters:
 *    in const char *dqb_path	- Path to dqb file
 *    in struct dq_blk *dq		- Quota structure for append to dqb
 * Return:
 *    0					- Ok
 *    EINVAL			- NULL ptr to dqb
 *    EACCES			- File not found (see errno for more detail)
 *    EACCES			- Bad rights
 *    EIO				- Can't open
 *    EIO				- Bad write
 *    EIO				- Can't close
 */
static int dq_append(in const char *dqb_path, in struct dq_blk *dq) {

	int				fd, status, cnt;

	// Check input params
	if ((0 == dqb_path) || (0 == dq)) return EINVAL;
	// Read counter for dqb
	if (0 != (status = read_dqb_cnt(dqb_path, &cnt))) return status;
	// open
	dq_open(fd, O_WRONLY | O_APPEND, dqb_path);
    // append
	dq_write(fd, dq, sizeof(struct dq_blk), dqb_path);
	// close
	dq_close(fd, dqb_path);
	// Write new counter for dqb
    cnt++;
	if (0 != (status = write_dqb_cnt(dqb_path, cnt))) return status;

	return 0;
}
/*
 * Add to dqb list quotas
 * Parameters:
 *    in const char		*dqb_path	- path to dqb file
 *    in struct dq_blk	*p		    - ptr to quota
 * Return:
 */
static int add_dqb(in const char *dqb_path, in struct dq_blk *p) {

	int				status;
	struct dq_blk	*dq, *r;
	__u32			cnt, ver;

    // Check input params 
    if ((0 == dqb_path) || (0 == p)) return -1;
    // Read ASP Disk Quota data base
	if (0 != (status = read_dqb(dqb_path, &dq, &cnt, &ver))) return status;
    // Find appropriate quota
    if (0 != (status = find_quotas_table(dq, cnt, p, &r))) {
        // Release memory
        free(dq);
        status = dq_append(dqb_path, p);
        return status;
    }
    // Set quota fields
    quota_set(r, p);
    // Reset REMOVE and DISABLE flags
	r->flags &= ~(DQ_ASP_REMOVE | DQ_ASP_DISABLE);
	// If hit to table - flush table on the disk
    status = flush_tbl2dqb(dqb_path, dq, cnt);
    // Release memory
    free(dq);
    return status;
}
