/*
*
* Author:              Vadim V. Sinitsyn
* E-mail:              sinvv@asplinux.ru
*
* $Header: /home/cvs/ASPcomplete/utils/quota_off.c,v 1.6 2000/10/17 17:09:02 sinvv Exp $
*
* Remarks:
*
* Last correct: 17/Oct/00
*
*/
#include <stdio.h>
#include <pwd.h>
#include <sys/types.h>
#include "dqlib.h"

static void		usage(in const char *task);

static struct passwd    *pwd;
static int	n_ve, type, ret, id=0;
static int	a_flag=0, 
			v_flag=0,
			u_flag=0,
			g_flag=0,
            ua_flag=0,ga_flag=0,va_flag=0,
			d_flag=0;

int main(int argc, char **argv) {

    struct dq_blk dq = { 0 };

    // Check policy
	check_root;
	// Set ve number
	if (0 == is_ve0()) n_ve = VEID_0; else n_ve = my_veid;
    // Kernel compile with ASP Disk Quota?
    DQ_ASP_IS;

	while ((ret = getopt(argc, argv, "hav:u:g:VUGD")) != EOF) {
		switch ( ret ) {
			case 'h':
			case '?':
					usage(argv[0]);
					return 1;
			case 'a':
					a_flag = 1;
					type = -1; 
					break;
			case 'u':
					u_flag = 1;
					type = DQ_ASP_UID;
                    // Check string: only decimal digits?
                    if (str_isdigit(optarg)) {
                        // Try found user with mnemonic name
                        pwd = getpwnam(optarg);
                        if (0 == pwd) {
                            printf("User %s not exist\n", optarg);
                            exit(1);
                        }
                        // Success
                        id = pwd->pw_uid;
                    }
                    else sscanf(optarg,"%u", &id);
                    dq.uid = id;
                    break;
			case 'g':
					g_flag = 1;
					type = DQ_ASP_GID;
                    sscanf(optarg,"%u", &id);
                    dq.gid = id;
					break;
			case 'v':
					if (VEID_0 != n_ve) usage(argv[0]);
					v_flag = 1;
					type = DQ_ASP_VEID;
                    sscanf(optarg,"%u", &id);
					break;
			case 'U': 
					ua_flag = 1;
                    type = DQ_ASP_UID | DQ_ASP_GLOBAL;
					break;
			case 'G': 
					ga_flag = 1;
                    type = DQ_ASP_GID | DQ_ASP_GLOBAL;
					break;
			case 'V': 
					// VEroot not allowed operation with VE!
					if (VEID_0 != n_ve) usage(argv[0]);
					va_flag = 1;
                    type = DQ_ASP_VEID | DQ_ASP_GLOBAL;
					break;
            case 'D':
                    d_flag = 1;
                    break;
		}
	}
	
    // Debug print
    if ( d_flag ) {
        printf("id=%d;a_flag=%d;u_flag=%d;g_flag=%d;v_flag=%d;ua_flag=%d;ga_flag=%d;va_flag=%d\n",
            id,a_flag,u_flag,g_flag,v_flag,ua_flag,ga_flag,va_flag);
    }

    // Set default value
    if (0 == (a_flag + u_flag + g_flag + v_flag + ua_flag + ga_flag + va_flag)) a_flag = 1; // default value
	// Check flags consistency
	if (1 != (a_flag + u_flag + g_flag + v_flag + ua_flag + ga_flag + va_flag)) usage(argv[0]);

    // Flush to disk
	dq_sync_u();

    if ( a_flag ) 
        return dq_asp(DQ_ASP_CMD_OFF, n_ve, 0, -1);
    else if (type & DQ_ASP_GLOBAL) 
        return dq_asp(DQ_ASP_CMD_OFF, n_ve, 0, type & DQ_ASP_VEGUID);

    dq.flags = type;
    dq.veid = n_ve;
    return dq_asp(DQ_ASP_CMD_OFF, 0, (__u32)&dq,0);
}
/*
 * Usage message
 * Parameters:
 *    in const char *task  - Task name
 * Return:
 *    None
 */
static void usage(in const char *task) {

	if (VEID_0 == n_ve) printf("Usage: %s [-havugVUG]\n", task);
	else printf("Usage: %s [-haugUG]\n", task);

	printf("\t-h\t\thelp for more information\n");

	printf("\t-a\t\tunload all quotas\n");

	printf("\t-U\t\tunload quotas for all users\n");
	printf("\t-G\t\tunload quotas for all groups\n");
	if (VEID_0 == n_ve) printf("\t-V\t\tunload quotas for all VEs\n");

	printf("\t-u n\t\tunload quota for users with uid number n\n");
	printf("\t-g n\t\tunload quota for groups with gid number n\n");
	if (VEID_0 == n_ve) printf("\t-v n\t\tunload quota for VE with veid number n\n");

	exit(0);
}
