/*
*
* Author:              Vadim V. Sinitsyn
* E-mail:              sinvv@asplinux.ru
*
* $Header: /home/cvs/ASPcomplete/utils/quota_on.c,v 1.11 2000/10/18 08:15:04 sinvv Exp $
*
* Remarks:
*
* Last correct: 18/Oct/00
*
*/
#include <stdio.h>
#include <string.h>
#include <pwd.h>
#include <sys/types.h>
#include "dqlib.h"

static void		usage(in const char *task);

static int quota_on(
    in const char *dqb_path, 
    in __u32 veid,
	in __u32 type,
	in __u32 flags,
    in __u32 filter_off);

static	path	dqb_path = DEFAULT_DQB;

static struct passwd    *pwd;
static int	n_ve, type, ret, id=0;
static int	a_flag=0,
			f_flag=0, 
			v_flag=0,
			u_flag=0,
			g_flag=0,
            ua_flag=0,ga_flag=0,va_flag=0,
            d_flag=0;

int main(int argc, char **argv) {

	// Check policy
	check_root;
	// Set ve number
	if (0 == is_ve0()) n_ve = VEID_0; else n_ve = my_veid;
    // Kernel compile with ASP Disk Quota?
    DQ_ASP_IS;

	while ((ret = getopt(argc, argv, "hf:av:u:g:VUGD")) != EOF) {
		switch ( ret ) {
			case 'h':
			case '?':
					usage(argv[0]);
					return 1;
			case 'f':
                    f_flag = 1;
					strncpy(dqb_path, optarg, PATH_SIZE-1);
					dqb_path[PATH_SIZE-1] = 0;
					break;
			case 'a':
					a_flag = 1;
					type = -1; 
					break;
			case 'u':
					u_flag = 1;
					type = DQ_ASP_UID;
                    // Check string: only decimal digits?
                    if (str_isdigit(optarg)) {
                        // Try found user with mnemonic name
                        pwd = getpwnam(optarg);
                        if (0 == pwd) {
                            printf("User %s not exist\n", optarg);
                            exit(1);
                        }
                        // Success
                        id = pwd->pw_uid;
                    }
                    else sscanf(optarg,"%u", &id);
					break;
			case 'g':
					g_flag = 1;
					type = DQ_ASP_GID;
                    sscanf(optarg,"%u", &id);
					break;
			case 'v':
					if (VEID_0 != n_ve) usage(argv[0]);
					v_flag = 1;
					type = DQ_ASP_VEID;
                    sscanf(optarg,"%u", &id);
					break;
			case 'U': 
					ua_flag = 1;
                    type = DQ_ASP_UID | DQ_ASP_GLOBAL;
					break;
			case 'G': 
					ga_flag = 1;
                    type = DQ_ASP_GID | DQ_ASP_GLOBAL;
					break;
			case 'V': 
					// VEroot not allowed operation with VE!
					if (VEID_0 != n_ve) usage(argv[0]);
					va_flag = 1;
                    type = DQ_ASP_VEID | DQ_ASP_GLOBAL;
					break;
            case 'D':
                    d_flag = 1;
                    break;
		}
	}

    // Debug print
    if ( d_flag ) {
        printf("id=%d;a_flag=%d;u_flag=%d;g_flag=%d;v_flag=%d;ua_flag=%d;ga_flag=%d;va_flag=%d\n",
            id,a_flag,u_flag,g_flag,v_flag,ua_flag,ga_flag,va_flag);
    }
    // Set default value
    if (0 == (a_flag + u_flag + g_flag + v_flag + ua_flag + ga_flag + va_flag)) a_flag = 1; // default value
	// Check flags consistency
	if (1 != (a_flag + u_flag + g_flag + v_flag + ua_flag + ga_flag + va_flag)) usage(argv[0]);

    // Set dqb_path
    if (0 == f_flag) {
        // Path to data base not defined explicitly 
        char *p;
        if (0 != (p = getenv(PATH_DQB_ENV))) {
            strcpy(dqb_path, p);
            if ( d_flag ) printf("dqb_path=%s\n", dqb_path);
        }
    }

    // Flush to disk
	dq_sync_u();

    return quota_on(dqb_path, n_ve, id, type, a_flag);
}
/*
 * Usage message
 * Parameters:
 *    in const char *task  - Task name
 * Return:
 *    None
 */
static void usage(in const char *task) {

	if (VEID_0 == n_ve) printf("Usage: %s [-hfavugVUG]\n", task);
	else printf("Usage: %s [-hfaugUG]\n", task);

	printf("\t-h\t\thelp for more information\n");
	printf("\t-f\t\tpath to dqb file (default: /var/quota/dqb)\n");

	printf("\t-a\t\tload all quotas\n");

	printf("\t-U\t\tload quotas for all users\n");
	printf("\t-G\t\tload quotas for all groups\n");
	if (VEID_0 == n_ve) printf("\t-V\t\tload quotas for all VEs\n");

	printf("\t-u n\t\tload quota for users with uid number n\n");
	printf("\t-g n\t\tload quota for groups with gid number n\n");
	if (VEID_0 == n_ve) printf("\t-v n\t\tload quota for VE with veid number n\n");

	exit(0);
}
/*
 * Load quotas
 * Parameters:
 *    in const char *dqb_path	- Path to dqb file
 *    in __u32 veid		        - Work veid
 *    in __u32 n			    - Number for filter 
 *    in __u32 flags			- Flags for filter
 *    in __u32 filter_off		- Filter on/off
 * Return:
 *    0					- Ok
 *    EINVAL			- NULL ptr to dqb
 *    EACCES			- File not found (see errno for more detail)
 *    EACCES			- Bad rights
 *    EIO				- Can't open
 *    EIO				- Bad write
 *    EIO				- Bad seek
 *    EIO				- Can't close
 */
static int quota_on(
    in const char *dqb_path, 
    in __u32 veid,
	in __u32 n,
	in __u32 flags,
    in __u32 filter_off) {

	int				status;
	__u32			res, i, ver, cnt;
	struct dq_blk	*p, *q;

	// Check input params
	if (NULL == dqb_path) return EINVAL;
	// Read dqb
	if (0 != (status = read_dqb(dqb_path, &p, &cnt, &ver))) return status;

    for (i = 0, q = p; i < cnt; i++, q++) {
		// If mark to delete - ignore
		if (DQ_ASP_IS_REMOVE(q->flags)) continue;

        if ( filter_off ) res = 1;
        else {
		    res = 0;
			switch (flags & (DQ_ASP_VEGUID | DQ_ASP_GLOBAL)) {
                case DQ_ASP_VEID:
					if ((n == q->veid) && (VEID_0 == n_ve) && (DQ_ASP_VEID == q->flags)) res = 1;
                    break;
                case DQ_ASP_UID:
                    if ((q->uid == n) && (q->veid == n_ve) && (DQ_ASP_UID == q->flags)) res = 1;
                    break;
                case DQ_ASP_GID:
                    if ((q->gid == n) && (q->veid == n_ve) && (DQ_ASP_GID == q->flags)) res = 1;
                    break;
				case (DQ_ASP_VEID | DQ_ASP_GLOBAL):
					if ((VEID_0 == n_ve) && (DQ_ASP_VEID == q->flags)) res = 1;
                    break;
				case (DQ_ASP_UID | DQ_ASP_GLOBAL):
					if ((q->veid == n_ve) && (DQ_ASP_UID == q->flags)) res = 1;
                    break;
				case (DQ_ASP_GID | DQ_ASP_GLOBAL):
					if ((q->veid == n_ve) && (DQ_ASP_GID == q->flags)) res = 1;
                    break;
            }
        }
        // Put quota to kernel
		if (0 != res) dq_put_index(q, dqb_path, i);
	}
	return 0;
}

