/*
*
* Author:              Vadim V. Sinitsyn
* E-mail:              sinvv@asplinux.ru
*
* $Header: /home/cvs/ASPcomplete/utils/quotachk.c,v 1.3 2000/10/17 17:09:32 sinvv Exp $
*
* Remarks:
*
* Last correct: 17/Oct/00
*
*/
#include <stdio.h>
#include <string.h>
#include "dqlib.h"

static void	usage(in const char *task);

path		quota_path;
				
int		n_ve, id, ret, status;
int		p_flag=0, 
		u_flag=0, 
		g_flag=0, 
		v_flag=0,
		d_flag=0;

__u32   type, units, inodes;

struct dq_blk   q;

int main(int argc, char **argv) {

    // Check policy
	check_root;
	// Set ve number
	if (0 == is_ve0()) n_ve = VEID_0; else n_ve = my_veid;
    // Kernel compile with ASP Disk Quota?
    DQ_ASP_IS;

	while ((ret = getopt(argc, argv, "hp:v:u:g:D")) != EOF) {
		switch ( ret ) {
			case 'h':
			case '?':
					usage(argv[0]);
					return 1;
			case 'p':
					p_flag = 1;
					strncpy(quota_path, optarg, PATH_SIZE-1);
					quota_path[PATH_SIZE-1] = 0;
					break;
			case 'u':
					u_flag = 1;
					type = DQ_ASP_UID;
                    sscanf(optarg, "%d", &id);
					break;
			case 'g':
					g_flag = 1;
					type = DQ_ASP_GID;
                    sscanf(optarg, "%d", &id);
					break;
			case 'v':
					v_flag = 1;
					type = DQ_ASP_VEID;
                    sscanf(optarg, "%d", &id);
					break;
            case 'D':
                    d_flag = 1;
                    break;
		}
	}

    // Debug print
    if ( d_flag ) {
        printf("p_flag=%d;u_flag=%d;g_flag=%d;v_flag=%d\n",
            p_flag,u_flag,g_flag,v_flag);
    }

    // Check flags
    if (0 == p_flag) usage(argv[0]);
	if (1 != (v_flag + u_flag + g_flag)) usage(argv[0]);

    if (0 == (status = dir_size(quota_path, type, id , 1, &units, &inodes))) {
        printf("for path=%s: units=%u; inodes=%u\n", quota_path, units, inodes);
    }
    else {
		printf("dir_size error: status=%d\n", status);
		return 1;
	}

    DQ_ASP_SET_DQ_BLK(&q,id,type);

    if (0 != (status = dq_asp(DQ_ASP_CMD_GET, (__u32)&q, 0, 0))) {
        printf("syscall dq error: status=%d\n", status);
        return 1;
    }

    if ((units != q.unit_current) || (inodes != q.inode_current)) {
        printf("Error: initquota/dqb: units=%u/%u; inodes=%u/%u\n", 
            units, q.unit_current, inodes, q.inode_current);

    }

    return 0;
}
/*
 * Usage message
 * Parameters:
 *    in const char *task  - Task name
 * Return:
 *    None
 */
static void usage(in const char *task) {

	if (VEID_0 == n_ve) printf("Usage: %s [-hpvug]\n", task);
	else printf("Usage: %s [-hpug]\n", task);

	printf("\t-h\t\thelp for more information\n");
	printf("\t-p quota_path\tpath to check quota\n");
	printf("\t-u id\t\tuid for check\n");
	printf("\t-g id\t\tgid for check\n");
	if (VEID_0 == n_ve) printf("\t-v id\t\tveid for check\n");
	exit(1);
}
