/*
*
* Author:              Vadim V. Sinitsyn
* E-mail:              sinvv@asplinux.ru
*
* $Header: /home/cvs/ASPcomplete/utils/repquota.c,v 1.12 2000/10/18 08:15:11 sinvv Exp $
*
* Remarks:
*
* Last correct: 18/Oct/00
*
*/
#include <stdio.h>
#include <string.h>
#include "dqlib.h"

static void	usage(in const char *task);
static int	print_dqb(in const char *dqb_path, in __u32 n, in __u32 flags, in __u32 filter_off, in __u32 unit_out_format);
/*
* groot:
*	1. uid-type - user from veid 0 (former)
*	2. gid-type - group from veid 0 (former)
*	3. veid-type - (!=0-!!!) - for appropriate VE
* back door:
*	4: if designate uid or/and gid - groot possible create private (VE) restriction,
*	namely (veid,uid,gid) (veid!=0)
* veroot
*	1. uid-type - user from native veid (veid !=0)
*	2. gid-type - group from native veid (veid !=0)
*	3. veid-type - must be reject
*/
static	path	dqb_path = DEFAULT_DQB;

static	int f_flag=0, 
			a_flag=0,
			c_flag=0,d_flag=0,
            u_flag=0,g_flag=0,v_flag=0,
            ua_flag=0,ga_flag=0,va_flag=0,
			Kb_flag=0, Mb_flag=0, 
            unit_out_format=0;

static int	n_ve, id=0;

int main(int argc, char **argv) {

	int				ret;
	int			type=0;

	// Check policy
	check_root;
	// Set ve number
	if (0 == is_ve0()) n_ve = VEID_0; else n_ve = my_veid;
    // Kernel compile with ASP Disk Quota?
    DQ_ASP_IS;

	while ((ret = getopt(argc, argv, "hf:aVGUv:u:g:kKmMcD")) != EOF) {
		switch ( ret ) {
			case 'h':
			case '?':
					usage(argv[0]);
					return 1;
			case 'c':
					c_flag = 1;
					break;
			case 'k':
			case 'K':
					Kb_flag = 1;
					unit_out_format = DQ_ASP_FORMAT_UNIT_KB;
					break;
			case 'm':
			case 'M':
					Mb_flag = 1;
					unit_out_format = DQ_ASP_FORMAT_UNIT_MB;
					break;
			case 'f': 
					f_flag = 1; 
					strncpy(dqb_path, optarg, PATH_SIZE-1);
					dqb_path[PATH_SIZE-1] = 0;
					break;
			case 'a':
					a_flag = 1;
					break;
			case 'U': 
					ua_flag = 1;
                    type = DQ_ASP_UID | DQ_ASP_GLOBAL;
					break;
			case 'G': 
					ga_flag = 1;
                    type = DQ_ASP_GID | DQ_ASP_GLOBAL;
					break;
			case 'V': 
					// VEroot not allowed operation with VE!
					if (VEID_0 != n_ve) usage(argv[0]);
					va_flag = 1;
                    type = DQ_ASP_VEID | DQ_ASP_GLOBAL;
					break;
			case 'u': 
					u_flag = 1;
                    type = DQ_ASP_UID;
                    sscanf(optarg,"%u", &id);
					break;
			case 'g': 
					g_flag = 1;
                    type = DQ_ASP_GID;
                    sscanf(optarg,"%u", &id);
					break;
			case 'v': 
					// VEroot not allowed operation with VE!
					if (VEID_0 != n_ve) usage(argv[0]);
					v_flag = 1;
                    type = DQ_ASP_VEID;
                    sscanf(optarg,"%u", &id);
					break;
            case 'D':
                    d_flag = 1;
                    break;
		}
	}

    // Debug print
    if ( d_flag ) {
        printf("c_flag=%d;a_flag=%d;u_flag=%d;g_flag=%d;v_flag=%d;ua_flag=%d;ga_flag=%d;va_flag=%d;Kb_flag=%d;Mb_flag=%d;f_flag=%d;dqb_path=%s;id=%d\n",
        c_flag,a_flag,u_flag,g_flag,v_flag,ua_flag,ga_flag,va_flag,Kb_flag,Mb_flag,f_flag,dqb_path,id);
    }
    // Set default value
    if (0 == (a_flag + u_flag + g_flag + v_flag + ua_flag + ga_flag + va_flag)) a_flag = 1; // default value
	// Check flags consistency
	if (1 != (a_flag + u_flag + g_flag + v_flag + ua_flag + ga_flag + va_flag)) usage(argv[0]);

    // Check flag consistency
	if (1 < (Kb_flag + Mb_flag)) usage(argv[0]);

    // Set dqb_path
    if (0 == f_flag) {
        // Path to data base not defined explicitly 
        char *p;
        if (0 != (p = getenv(PATH_DQB_ENV))) {
            strcpy(dqb_path, p);
            if ( d_flag ) printf("dqb_path=%s\n", dqb_path);
        }
    }

    // Flush to disk
	dq_sync_u();

    // Print data
	print_dqb(dqb_path, id, type, a_flag, unit_out_format);

	return 0;
}
/*
 * Usage message
 * Parameters:
 *    in const char *task  - Task name
 * Return:
 *    None
 */
static void usage(in const char *task) {

	if (VEID_0 == n_ve) printf("Usage: %s [-favugVUG]\n", task);
	else printf("Usage: %s [-augUG]\n", task);

	printf("\t-a\t\tprint all quotas\n");
	if (VEID_0 == n_ve) printf("\t-f path\t\tpath to dqb file (default: /var/quota/dqb)\n");

	printf("\t-U\t\tprint quotas for all users\n");
	printf("\t-G\t\tprint quotas for all groups\n");
	if (VEID_0 == n_ve) printf("\t-V\t\tprint quotas for all VEs\n");

	printf("\t-u n\t\tprint quota for users with uid number n\n");
	printf("\t-g n\t\tprint quota for groups with gid number n\n");
	if (VEID_0 == n_ve) printf("\t-v n\t\tprint quota for VE with veid number n\n");
	
	printf("\t-k (or -K)\tprint quota units in Kbytes (default: in bytes)\n");
	printf("\t-m (or -M)\tprint quota units in Mbytes (default: in bytes)\n");
	exit(0);
}
/*
 * Print dqb content
 * Parameters:
 *    in const char *dqb_path	- Path to quota data base file
 *    in __u32 n					- Number for filter 
 *    in __u32 flags				- Flags for filter
 *    in __u32 filter_off			- Filter on/off
 *    in __u32 unit_out_format	- Output format for unit (Bytes/Kb/Mb)
 * Return:
 *    0					- Ok
 *    EINVAL			- NULL ptr to dqb
 *    ESRCH				- File not found (see errno for more detail)
 *    EACCES			- Bad rights
 *    EIO				- Can't open/read/seek or close
 */
static int print_dqb(
	in const char *dqb_path, 
	in __u32 n,
	in __u32 flags,
	in __u32 filter_off,
	in __u32 unit_out_format) {

	int				status;
	__u32			cnt, ver, i, cnt_record=0;
	struct dq_blk	*dqb, *p, *q, dhash;

	// Check input params
	if (NULL == dqb_path) return EINVAL;

    // Read ASP Disk Quota data base
	if (0 != (status = read_dqb(dqb_path, &dqb, &cnt, &ver))) {
		printf("Can't read dqb %s: status error=%d\n", dqb_path, status);	
		return status;
	}

	for (i = 0, p = dqb; i < cnt; i++, p++) {

        // If set shadow flag '-c' - get data from RAM (not disk!) - for debug target
		if (0 != c_flag) {
			memcpy((void *)&dhash, (const void *)p, sizeof(struct dq_blk));
			if (0 != (status = dq_asp(DQ_ASP_CMD_GET, (__u32)&dhash, 0, 0))) {
				//printf("Bad DQ_ASP_CMD_GET: status=%d\n", status);
				continue;
			}
			else q = &dhash;
		}
		else q = p;

		// If mark to delete - ignore
		if (DQ_ASP_IS_REMOVE(q->flags)) continue;

		if ( filter_off ) {
			if (0 == cnt_record) print_dqblk_header();
			++cnt_record;
			print_dqblk(q, DQ_ASP_FORMAT_LINE, unit_out_format);
		}
		else {

			switch (flags & (DQ_ASP_VEGUID | DQ_ASP_GLOBAL)) {
				case DQ_ASP_VEID:
					if ((q->veid == n) && (VEID_0 == n_ve) && (DQ_ASP_VEID == q->flags)) {
						if (0 == cnt_record) print_dqblk_header();
						++cnt_record;
						print_dqblk(q, DQ_ASP_FORMAT_LINE, unit_out_format);
					}
					break;
				case DQ_ASP_UID:
					if ((q->uid == n) && (q->veid == n_ve) && (DQ_ASP_UID == q->flags)) {
						if (0 == cnt_record) print_dqblk_header();
						++cnt_record;
						print_dqblk(q, DQ_ASP_FORMAT_LINE, unit_out_format);
					}
					break;
				case DQ_ASP_GID:
					if ((q->gid == n) && (q->veid == n_ve) && (DQ_ASP_GID == q->flags)) {
						if (0 == cnt_record) print_dqblk_header();
						++cnt_record;
						print_dqblk(q, DQ_ASP_FORMAT_LINE, unit_out_format);
					}
					break;
				case (DQ_ASP_VEID | DQ_ASP_GLOBAL):
					if ((VEID_0 == n_ve) && (DQ_ASP_VEID == q->flags)) {
						if (0 == cnt_record) print_dqblk_header();
						++cnt_record;
						print_dqblk(q, DQ_ASP_FORMAT_LINE, unit_out_format);
					}
    				break;
				case (DQ_ASP_UID | DQ_ASP_GLOBAL):
					if ((q->veid == n_ve) && (DQ_ASP_UID == q->flags)) {
					    if (0 == cnt_record) print_dqblk_header();
					    ++cnt_record;
					    print_dqblk(q, DQ_ASP_FORMAT_LINE, unit_out_format);
                    }
					break;
				case (DQ_ASP_GID | DQ_ASP_GLOBAL):
					if ((q->veid == n_ve) && (DQ_ASP_GID == q->flags)) {
						if (0 == cnt_record) print_dqblk_header();
						++cnt_record;
						print_dqblk(q, DQ_ASP_FORMAT_LINE, unit_out_format);
					}
					break;
                default: break;
			}
		}
	}

    // Free table
	dq_free(dqb);
	return 0;
}
